"""
Tests for our mock_open helper
"""

import errno
import logging
import textwrap

import salt.utils.data
import salt.utils.files
import salt.utils.stringutils
from tests.support.mock import mock_open, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class MockOpenMixin:
    def _get_values(self, binary=False, multifile=False, split=False):
        if split:
            questions = (
                self.questions_bytes_lines if binary else self.questions_str_lines
            )
            answers = self.answers_bytes_lines if binary else self.answers_str_lines
        else:
            questions = self.questions_bytes if binary else self.questions_str
            answers = self.answers_bytes if binary else self.answers_str
        mode = "rb" if binary else "r"
        if multifile:
            read_data = self.contents_bytes if binary else self.contents
        else:
            read_data = self.questions_bytes if binary else self.questions
        return questions, answers, mode, read_data

    def _test_read(self, binary=False, multifile=False):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                result = self.fh.read()
                assert result == questions, result

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    result = self.fh2.read()
                    assert result == answers, result

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    result = self.fh3.read()
                    assert result == answers, result

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No patterns should have matched")
                except OSError:
                    # An IOError is expected here
                    pass

    def _test_read_explicit_size(self, binary=False, multifile=False):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                # Read 10 bytes
                result = self.fh.read(10)
                assert result == questions[:10], result
                # Read another 10 bytes
                result = self.fh.read(10)
                assert result == questions[10:20], result
                # Read the rest
                result = self.fh.read()
                assert result == questions[20:], result

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    # Read 10 bytes
                    result = self.fh2.read(10)
                    assert result == answers[:10], result
                    # Read another 10 bytes
                    result = self.fh2.read(10)
                    assert result == answers[10:20], result
                    # Read the rest
                    result = self.fh2.read()
                    assert result == answers[20:], result

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    # Read 10 bytes
                    result = self.fh3.read(10)
                    assert result == answers[:10], result
                    # Read another 10 bytes
                    result = self.fh3.read(10)
                    assert result == answers[10:20], result
                    # Read the rest
                    result = self.fh3.read()
                    assert result == answers[20:], result

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No globs should have matched")
                except OSError:
                    # An IOError is expected here
                    pass

    def _test_read_explicit_size_larger_than_file_size(
        self, binary=False, multifile=False
    ):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                result = self.fh.read(999999)
                assert result == questions, result

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    result = self.fh2.read(999999)
                    assert result == answers, result

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    result = self.fh3.read(999999)
                    assert result == answers, result

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No globs should have matched")
                except OSError:
                    # An IOError is expected here
                    pass

    def _test_read_for_loop(self, binary=False, multifile=False):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile, split=True
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                index = 0
                for line in self.fh:
                    assert line == questions[index], f"Line {index}: {line}"
                    index += 1

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    index = 0
                    for line in self.fh2:
                        assert line == answers[index], f"Line {index}: {line}"
                        index += 1

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    index = 0
                    for line in self.fh3:
                        assert line == answers[index], f"Line {index}: {line}"
                        index += 1

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No globs should have matched")
                except OSError:
                    # An IOError is expected here
                    pass

    def _test_read_readline(self, binary=False, multifile=False):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile, split=True
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                size = 8
                result = self.fh.read(size)
                assert result == questions[0][:size], result
                # Use .readline() to read the remainder of the line
                result = self.fh.readline()
                assert result == questions[0][size:], result
                # Read and check the other two lines
                result = self.fh.readline()
                assert result == questions[1], result
                result = self.fh.readline()
                assert result == questions[2], result

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    size = 20
                    result = self.fh2.read(size)
                    assert result == answers[0][:size], result
                    # Use .readline() to read the remainder of the line
                    result = self.fh2.readline()
                    assert result == answers[0][size:], result
                    # Read and check the other two lines
                    result = self.fh2.readline()
                    assert result == answers[1], result
                    result = self.fh2.readline()
                    assert result == answers[2], result

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    size = 20
                    result = self.fh3.read(size)
                    assert result == answers[0][:size], result
                    # Use .readline() to read the remainder of the line
                    result = self.fh3.readline()
                    assert result == answers[0][size:], result
                    # Read and check the other two lines
                    result = self.fh3.readline()
                    assert result == answers[1], result
                    result = self.fh3.readline()
                    assert result == answers[2], result

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No globs should have matched")
                except OSError:
                    # An IOError is expected here
                    pass

    def _test_readline_readlines(self, binary=False, multifile=False):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile, split=True
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                # Read the first line
                result = self.fh.readline()
                assert result == questions[0], result
                # Use .readlines() to read the remainder of the file
                result = self.fh.readlines()
                assert result == questions[1:], result

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    # Read the first line
                    result = self.fh2.readline()
                    assert result == answers[0], result
                    # Use .readlines() to read the remainder of the file
                    result = self.fh2.readlines()
                    assert result == answers[1:], result

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    # Read the first line
                    result = self.fh3.readline()
                    assert result == answers[0], result
                    # Use .readlines() to read the remainder of the file
                    result = self.fh3.readlines()
                    assert result == answers[1:], result

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No globs should have matched")
                except OSError:
                    # An IOError is expected here
                    pass

    def _test_readlines_multifile(self, binary=False, multifile=False):
        questions, answers, mode, read_data = self._get_values(
            binary=binary, multifile=multifile, split=True
        )

        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)):
            with salt.utils.files.fopen("foo.txt", mode) as self.fh:
                result = self.fh.readlines()
                assert result == questions, result

            if multifile:
                with salt.utils.files.fopen("bar.txt", mode) as self.fh2:
                    result = self.fh2.readlines()
                    assert result == answers, result

                with salt.utils.files.fopen("baz.txt", mode) as self.fh3:
                    result = self.fh3.readlines()
                    assert result == answers, result

                try:
                    with salt.utils.files.fopen("helloworld.txt"):
                        raise Exception("No globs should have matched")
                except OSError:
                    # An IOError is expected here
                    pass


class MockOpenTestCase(TestCase, MockOpenMixin):
    """
    Tests for our mock_open helper to ensure that it behaves as closely as
    possible to a real filehandle.
    """

    # Cyrllic characters used to test unicode handling
    questions = textwrap.dedent(
        """\
        Шнат is your name?
        Шнат is your quest?
        Шнат is the airspeed velocity of an unladen swallow?
        """
    )

    answers = textwrap.dedent(
        """\
        It is Аятнця, King of the Britons.
        To seek тне Holy Grail.
        Шнат do you mean? An African or European swallow?
        """
    )

    @classmethod
    def setUpClass(cls):
        cls.questions_lines = cls.questions.splitlines(True)
        cls.answers_lines = cls.answers.splitlines(True)

        cls.questions_str = salt.utils.stringutils.to_str(cls.questions)
        cls.answers_str = salt.utils.stringutils.to_str(cls.answers)
        cls.questions_str_lines = cls.questions_str.splitlines(True)
        cls.answers_str_lines = cls.answers_str.splitlines(True)

        cls.questions_bytes = salt.utils.stringutils.to_bytes(cls.questions)
        cls.answers_bytes = salt.utils.stringutils.to_bytes(cls.answers)
        cls.questions_bytes_lines = cls.questions_bytes.splitlines(True)
        cls.answers_bytes_lines = cls.answers_bytes.splitlines(True)

        # When this is used as the read_data, Python 2 should normalize
        # cls.questions and cls.answers to str types.
        cls.contents = {"foo.txt": cls.questions, "b*.txt": cls.answers}
        cls.contents_bytes = {
            "foo.txt": cls.questions_bytes,
            "b*.txt": cls.answers_bytes,
        }

        cls.read_data_as_list = [
            "foo",
            "bar",
            "спам",
            IOError(errno.EACCES, "Permission denied"),
        ]
        cls.normalized_read_data_as_list = salt.utils.data.decode(
            cls.read_data_as_list, to_str=True
        )
        cls.read_data_as_list_bytes = salt.utils.data.encode(cls.read_data_as_list)

    def tearDown(self):
        """
        Each test should read the entire contents of the mocked filehandle(s).
        This confirms that the other read functions return empty strings/lists,
        to simulate being at EOF.
        """
        for handle_name in ("fh", "fh2", "fh3"):
            try:
                fh = getattr(self, handle_name)
            except AttributeError:
                continue
            log.debug("Running tearDown tests for self.%s", handle_name)
            try:
                result = fh.read(5)
                assert not result, result
                result = fh.read()
                assert not result, result
                result = fh.readline()
                assert not result, result
                result = fh.readlines()
                assert not result, result
                # Last but not least, try to read using a for loop. This should not
                # read anything as we should hit EOF immediately, before the generator
                # in the mocked filehandle has a chance to yield anything. So the
                # exception will only be raised if we aren't at EOF already.
                for line in fh:
                    raise Exception(
                        "Instead of EOF, read the following from {}: {}".format(
                            handle_name, line
                        )
                    )
            except OSError as exc:
                if str(exc) != "File not open for reading":
                    raise
            del fh

    def test_read(self):
        """
        Test reading the entire file
        """
        self._test_read(binary=False, multifile=False)
        self._test_read(binary=True, multifile=False)
        self._test_read(binary=False, multifile=True)
        self._test_read(binary=True, multifile=True)

    def test_read_explicit_size(self):
        """
        Test reading with explicit sizes
        """
        self._test_read_explicit_size(binary=False, multifile=False)
        self._test_read_explicit_size(binary=True, multifile=False)
        self._test_read_explicit_size(binary=False, multifile=True)
        self._test_read_explicit_size(binary=True, multifile=True)

    def test_read_explicit_size_larger_than_file_size(self):
        """
        Test reading with an explicit size larger than the size of read_data.
        This ensures that we just return the contents up until EOF and that we
        don't raise any errors due to the desired size being larger than the
        mocked file's size.
        """
        self._test_read_explicit_size_larger_than_file_size(
            binary=False, multifile=False
        )
        self._test_read_explicit_size_larger_than_file_size(
            binary=True, multifile=False
        )
        self._test_read_explicit_size_larger_than_file_size(
            binary=False, multifile=True
        )
        self._test_read_explicit_size_larger_than_file_size(binary=True, multifile=True)

    def test_read_for_loop(self):
        """
        Test reading the contents of the file line by line in a for loop
        """
        self._test_read_for_loop(binary=False, multifile=False)
        self._test_read_for_loop(binary=True, multifile=False)
        self._test_read_for_loop(binary=False, multifile=True)
        self._test_read_for_loop(binary=True, multifile=True)

    def test_read_readline(self):
        """
        Test reading part of a line using .read(), then reading the rest of the
        line (and subsequent lines) using .readline().
        """
        self._test_read_readline(binary=False, multifile=False)
        self._test_read_readline(binary=True, multifile=False)
        self._test_read_readline(binary=False, multifile=True)
        self._test_read_readline(binary=True, multifile=True)

    def test_readline_readlines(self):
        """
        Test reading the first line using .readline(), then reading the rest of
        the file using .readlines().
        """
        self._test_readline_readlines(binary=False, multifile=False)
        self._test_readline_readlines(binary=True, multifile=False)
        self._test_readline_readlines(binary=False, multifile=True)
        self._test_readline_readlines(binary=True, multifile=True)

    def test_readlines(self):
        """
        Test reading the entire file using .readlines
        """
        self._test_readlines_multifile(binary=False, multifile=False)
        self._test_readlines_multifile(binary=True, multifile=False)
        self._test_readlines_multifile(binary=False, multifile=True)
        self._test_readlines_multifile(binary=True, multifile=True)

    def test_read_data_converted_to_dict(self):
        """
        Test that a non-dict value for read_data is converted to a dict mapping
        '*' to that value.
        """
        contents = "спам"
        normalized = salt.utils.stringutils.to_str(contents)
        with patch("salt.utils.files.fopen", mock_open(read_data=contents)) as m_open:
            assert m_open.read_data == {"*": normalized}, m_open.read_data

        with patch(
            "salt.utils.files.fopen", mock_open(read_data=self.read_data_as_list)
        ) as m_open:
            assert m_open.read_data == {
                "*": self.normalized_read_data_as_list,
            }, m_open.read_data

    def test_read_data_list(self):
        """
        Test read_data when it is a list
        """
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=self.read_data_as_list)
        ):
            for value in self.normalized_read_data_as_list:
                try:
                    with salt.utils.files.fopen("foo.txt") as self.fh:
                        result = self.fh.read()
                        assert result == value, result
                except OSError:
                    # Only raise the caught exception if it wasn't expected
                    # (i.e. if value is not an exception)
                    if not isinstance(value, IOError):
                        raise

    def test_read_data_list_bytes(self):
        """
        Test read_data when it is a list and the value is a bytestring
        """
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=self.read_data_as_list_bytes)
        ):
            for value in self.read_data_as_list_bytes:
                try:
                    with salt.utils.files.fopen("foo.txt", "rb") as self.fh:
                        result = self.fh.read()
                        assert result == value, result
                except OSError:
                    # Only raise the caught exception if it wasn't expected
                    # (i.e. if value is not an exception)
                    if not isinstance(value, IOError):
                        raise

    def test_tell(self):
        """
        Test the implementation of tell
        """
        with patch("salt.utils.files.fopen", mock_open(read_data=self.contents)):
            # Try with reading explicit sizes and then reading the rest of the
            # file.
            with salt.utils.files.fopen("foo.txt") as self.fh:
                self.fh.read(5)
                loc = self.fh.tell()
                assert loc == 5, loc
                self.fh.read(12)
                loc = self.fh.tell()
                assert loc == 17, loc
                self.fh.read()
                loc = self.fh.tell()
                assert loc == len(self.questions_str), loc

            # Try reading way more content then actually exists in the file,
            # tell() should return a value equal to the length of the content
            with salt.utils.files.fopen("foo.txt") as self.fh:
                self.fh.read(999999)
                loc = self.fh.tell()
                assert loc == len(self.questions_str), loc

            # Try reading a few bytes using .read(), then the rest of the line
            # using .readline(), then the rest of the file using .readlines(),
            # and check the location after each read.
            with salt.utils.files.fopen("foo.txt") as self.fh:
                # Read a few bytes
                self.fh.read(5)
                loc = self.fh.tell()
                assert loc == 5, loc
                # Read the rest of the line. Location should then be at the end
                # of the first line.
                self.fh.readline()
                loc = self.fh.tell()
                assert loc == len(self.questions_str_lines[0]), loc
                # Read the rest of the file using .readlines()
                self.fh.readlines()
                loc = self.fh.tell()
                assert loc == len(self.questions_str), loc

            # Check location while iterating through the filehandle
            with salt.utils.files.fopen("foo.txt") as self.fh:
                index = 0
                for _ in self.fh:
                    index += 1
                    loc = self.fh.tell()
                    assert loc == sum(
                        len(x) for x in self.questions_str_lines[:index]
                    ), loc

    def test_write(self):
        """
        Test writing to a filehandle using .write()
        """
        # Test opening for non-binary writing
        with patch("salt.utils.files.fopen", mock_open()):
            with salt.utils.files.fopen("foo.txt", "w") as self.fh:
                for line in self.questions_str_lines:
                    self.fh.write(line)
                assert (
                    self.fh.write_calls == self.questions_str_lines
                ), self.fh.write_calls

        # Test opening for binary writing using "wb"
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "wb") as self.fh:
                for line in self.questions_bytes_lines:
                    self.fh.write(line)
                assert (
                    self.fh.write_calls == self.questions_bytes_lines
                ), self.fh.write_calls

        # Test opening for binary writing using "ab"
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "ab") as self.fh:
                for line in self.questions_bytes_lines:
                    self.fh.write(line)
                assert (
                    self.fh.write_calls == self.questions_bytes_lines
                ), self.fh.write_calls

        # Test opening for read-and-write using "r+b"
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "r+b") as self.fh:
                for line in self.questions_bytes_lines:
                    self.fh.write(line)
                assert (
                    self.fh.write_calls == self.questions_bytes_lines
                ), self.fh.write_calls

        # Test trying to write str types to a binary filehandle
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "wb") as self.fh:
                try:
                    self.fh.write("foo\n")
                except TypeError:
                    # This exception is expected on Python 3
                    pass
                else:
                    # This write should work fine on Python 2
                    raise Exception(
                        "Should not have been able to write a str to a "
                        "binary filehandle"
                    )

        # Test trying to write bytestrings to a non-binary filehandle
        with patch("salt.utils.files.fopen", mock_open()):
            with salt.utils.files.fopen("foo.txt", "w") as self.fh:
                try:
                    self.fh.write(b"foo\n")
                except TypeError:
                    # This exception is expected on Python 3
                    pass
                else:
                    # This write should work fine on Python 2
                    raise Exception(
                        "Should not have been able to write a bytestring "
                        "to a non-binary filehandle"
                    )

    def test_writelines(self):
        """
        Test writing to a filehandle using .writelines()
        """
        # Test opening for non-binary writing
        with patch("salt.utils.files.fopen", mock_open()):
            with salt.utils.files.fopen("foo.txt", "w") as self.fh:
                self.fh.writelines(self.questions_str_lines)
                assert self.fh.writelines_calls == [
                    self.questions_str_lines
                ], self.fh.writelines_calls

        # Test opening for binary writing using "wb"
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "wb") as self.fh:
                self.fh.writelines(self.questions_bytes_lines)
                assert self.fh.writelines_calls == [
                    self.questions_bytes_lines
                ], self.fh.writelines_calls

        # Test opening for binary writing using "ab"
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "ab") as self.fh:
                self.fh.writelines(self.questions_bytes_lines)
                assert self.fh.writelines_calls == [
                    self.questions_bytes_lines
                ], self.fh.writelines_calls

        # Test opening for read-and-write using "r+b"
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "r+b") as self.fh:
                self.fh.writelines(self.questions_bytes_lines)
                assert self.fh.writelines_calls == [
                    self.questions_bytes_lines
                ], self.fh.writelines_calls

        # Test trying to write str types to a binary filehandle
        with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
            with salt.utils.files.fopen("foo.txt", "wb") as self.fh:
                try:
                    self.fh.writelines(["foo\n"])
                except TypeError:
                    # This exception is expected on Python 3
                    pass
                else:
                    # This write should work fine on Python 2
                    raise Exception(
                        "Should not have been able to write a str to a "
                        "binary filehandle"
                    )

        # Test trying to write bytestrings to a non-binary filehandle
        with patch("salt.utils.files.fopen", mock_open()):
            with salt.utils.files.fopen("foo.txt", "w") as self.fh:
                try:
                    self.fh.write([b"foo\n"])
                except TypeError:
                    # This exception is expected on Python 3
                    pass
                else:
                    # This write should work fine on Python 2
                    raise Exception(
                        "Should not have been able to write a bytestring "
                        "to a non-binary filehandle"
                    )

    def test_open(self):
        """
        Test that opening a file for binary reading with string read_data
        fails, and that the same thing happens for non-binary filehandles and
        bytestring read_data.

        NOTE: This test should always pass on PY2 since MockOpen will normalize
        unicode types to str types.
        """
        try:
            with patch("salt.utils.files.fopen", mock_open()):
                try:
                    with salt.utils.files.fopen("foo.txt", "rb") as self.fh:
                        self.fh.read()
                except TypeError:
                    pass
                else:
                    raise Exception(
                        "Should not have been able open for binary read with "
                        "non-bytestring read_data"
                    )

            with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
                try:
                    with salt.utils.files.fopen("foo.txt", "r") as self.fh2:
                        self.fh2.read()
                except TypeError:
                    pass
                else:
                    raise Exception(
                        "Should not have been able open for non-binary read "
                        "with bytestring read_data"
                    )
        finally:
            # Make sure we destroy the filehandles before the teardown, as they
            # will also try to read and this will generate another exception
            delattr(self, "fh")
            delattr(self, "fh2")



import urllib.parse

import pytest

import salt.utils.json
import salt.utils.yaml
from salt.ext.tornado.httpclient import HTTPError
from salt.netapi.rest_tornado import saltnado

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
]


@pytest.fixture
def app_urls():
    return [
        ("/login", saltnado.SaltAuthHandler),
    ]


async def test_get(http_client):
    """
    We don't allow gets, so assert we get 401s
    """
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch("/login")
    assert exc.value.code == 401


async def test_login(
    http_client, content_type_map, auth_creds, subtests, client_config
):
    """
    Test valid logins
    """
    with subtests.test("Test in form encoded"):
        response = await http_client.fetch(
            "/login",
            method="POST",
            body=urllib.parse.urlencode(auth_creds),
            headers={"Content-Type": content_type_map["form"]},
        )

        cookies = response.headers["Set-Cookie"]
        assert response.code == 200
        response_obj = salt.utils.json.loads(response.body)["return"][0]
        token = response_obj["token"]
        assert f"session_id={token}" in cookies
        perms = response_obj["perms"]
        perms_config = client_config["external_auth"]["auto"][auth_creds["username"]]
        assert set(perms) == set(perms_config)
        assert "token" in response_obj  # TODO: verify that its valid?
        assert response_obj["user"] == auth_creds["username"]
        assert response_obj["eauth"] == auth_creds["eauth"]

    with subtests.test("Test in JSON"):
        response = await http_client.fetch(
            "/login",
            method="POST",
            body=salt.utils.json.dumps(auth_creds),
            headers={"Content-Type": content_type_map["json"]},
        )

        cookies = response.headers["Set-Cookie"]
        assert response.code == 200
        response_obj = salt.utils.json.loads(response.body)["return"][0]
        token = response_obj["token"]
        assert f"session_id={token}" in cookies
        perms = response_obj["perms"]
        perms_config = client_config["external_auth"]["auto"][auth_creds["username"]]
        assert set(perms) == set(perms_config)
        assert "token" in response_obj  # TODO: verify that its valid?
        assert response_obj["user"] == auth_creds["username"]
        assert response_obj["eauth"] == auth_creds["eauth"]

    with subtests.test("Test in YAML"):
        response = await http_client.fetch(
            "/login",
            method="POST",
            body=salt.utils.yaml.safe_dump(auth_creds),
            headers={"Content-Type": content_type_map["yaml"]},
        )

        cookies = response.headers["Set-Cookie"]
        assert response.code == 200
        response_obj = salt.utils.json.loads(response.body)["return"][0]
        token = response_obj["token"]
        assert f"session_id={token}" in cookies
        perms = response_obj["perms"]
        perms_config = client_config["external_auth"]["auto"][auth_creds["username"]]
        assert set(perms) == set(perms_config)
        assert "token" in response_obj  # TODO: verify that its valid?
        assert response_obj["user"] == auth_creds["username"]
        assert response_obj["eauth"] == auth_creds["eauth"]


async def test_login_missing_password(http_client, auth_creds, content_type_map):
    """
    Test logins with bad/missing passwords
    """
    bad_creds = []
    for key, val in auth_creds.items():
        if key == "password":
            continue
        bad_creds.append((key, val))
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/login",
            method="POST",
            body=urllib.parse.urlencode(bad_creds),
            headers={"Content-Type": content_type_map["form"]},
        )

    assert exc.value.code == 400


async def test_login_bad_creds(http_client, content_type_map, auth_creds):
    """
    Test logins with bad/missing passwords
    """
    bad_creds = []
    for key, val in auth_creds.items():
        if key == "username":
            val = val + "foo"
        if key == "eauth":
            val = "sharedsecret"
        bad_creds.append((key, val))

    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/login",
            method="POST",
            body=urllib.parse.urlencode(bad_creds),
            headers={"Content-Type": content_type_map["form"]},
        )

    assert exc.value.code == 401


async def test_login_invalid_data_structure(http_client, content_type_map, auth_creds):
    """
    Test logins with either list or string JSON payload
    """
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/login",
            method="POST",
            body=salt.utils.json.dumps(auth_creds),
            headers={"Content-Type": content_type_map["form"]},
        )

    assert exc.value.code == 400

    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/login",
            method="POST",
            body=salt.utils.json.dumps(42),
            headers={"Content-Type": content_type_map["form"]},
        )

    assert exc.value.code == 400

    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/login",
            method="POST",
            body=salt.utils.json.dumps("mystring42"),
            headers={"Content-Type": content_type_map["form"]},
        )

    assert exc.value.code == 400



"""
The function cache system allows for data to be stored on the master so it can be easily read by other minions
"""

import logging
import time
import traceback

import salt.channel.client
import salt.crypt
import salt.payload
import salt.transport
import salt.utils.args
import salt.utils.dictupdate
import salt.utils.event
import salt.utils.functools
import salt.utils.mine
import salt.utils.minions
import salt.utils.network
from salt.exceptions import SaltClientError

MINE_INTERNAL_KEYWORDS = frozenset(
    [
        "__pub_user",
        "__pub_arg",
        "__pub_fun",
        "__pub_jid",
        "__pub_tgt",
        "__pub_tgt_type",
        "__pub_ret",
    ]
)

__proxyenabled__ = ["*"]

log = logging.getLogger(__name__)


def _auth():
    """
    Return the auth object
    """
    if "auth" not in __context__:
        try:
            __context__["auth"] = salt.crypt.SAuth(__opts__)
        except SaltClientError:
            log.error(
                "Could not authenticate with master. Mine data will not be transmitted."
            )
    return __context__["auth"]


def _mine_function_available(func):
    if func not in __salt__:
        log.error("Function %s in mine_functions not available", func)
        return False
    return True


def _mine_send(load, opts):
    eventer = salt.utils.event.MinionEvent(opts, listen=False)
    event_ret = eventer.fire_event(load, "_minion_mine")
    # We need to pause here to allow for the decoupled nature of
    # events time to allow the mine to propagate
    time.sleep(0.5)
    return event_ret


def _mine_get(load, opts):
    with salt.channel.client.ReqChannel.factory(opts) as channel:
        return channel.send(load)


def _mine_store(mine_data, clear=False):
    """
    Helper function to store the provided mine data.
    This will store either locally in the cache (for masterless setups), or in
    the master's cache.

    :param dict mine_data: Dictionary with function_name: function_data to store.
    :param bool clear: Whether or not to clear (`True`) the mine data for the
        function names present in ``mine_data``, or update it (`False`).
    """
    # Store in the salt-minion's local cache
    if __opts__["file_client"] == "local":
        if not clear:
            old = __salt__["data.get"]("mine_cache")
            if isinstance(old, dict):
                old.update(mine_data)
                mine_data = old
        return __salt__["data.update"]("mine_cache", mine_data)
    # Store on the salt master
    load = {
        "cmd": "_mine",
        "data": mine_data,
        "id": __opts__["id"],
        "clear": clear,
    }
    return _mine_send(load, __opts__)


def update(clear=False, mine_functions=None):
    """
    Call the configured functions and send the data back up to the master.
    The functions to be called are merged from the master config, pillar and
    minion config under the option `mine_functions`:

    .. code-block:: yaml

        mine_functions:
          network.ip_addrs:
            - eth0
          disk.usage: []

    This function accepts the following arguments:

    :param bool clear: Default: ``False``
        Specifies whether updating will clear the existing values (``True``), or
        whether it will update them (``False``).

    :param dict mine_functions:
        Update (or clear, see ``clear``) the mine data on these functions only.
        This will need to have the structure as defined on
        https://docs.saltproject.io/en/latest/topics/mine/index.html#mine-functions

        This feature can be used when updating the mine for functions
        that require a refresh at different intervals than the rest of
        the functions specified under `mine_functions` in the
        minion/master config or pillar.
        A potential use would be together with the `scheduler`, for example:

        .. code-block:: yaml

            schedule:
              lldp_mine_update:
                function: mine.update
                kwargs:
                    mine_functions:
                      net.lldp: []
                hours: 12

        In the example above, the mine for `net.lldp` would be refreshed
        every 12 hours, while  `network.ip_addrs` would continue to be updated
        as specified in `mine_interval`.

    The function cache will be populated with information from executing these
    functions

    CLI Example:

    .. code-block:: bash

        salt '*' mine.update
    """
    if not mine_functions:
        mine_functions = __salt__["config.merge"]("mine_functions", {})
        # If we don't have any mine functions configured, then we should just bail out
        if not mine_functions:
            return
    elif isinstance(mine_functions, list):
        mine_functions = {fun: {} for fun in mine_functions}
    elif isinstance(mine_functions, dict):
        pass
    else:
        return

    mine_data = {}
    for function_alias, function_data in mine_functions.items():
        (
            function_name,
            function_args,
            function_kwargs,
            minion_acl,
        ) = salt.utils.mine.parse_function_definition(function_data)
        if not _mine_function_available(function_name or function_alias):
            continue
        try:
            res = salt.utils.functools.call_function(
                __salt__[function_name or function_alias],
                *function_args,
                **function_kwargs,
            )
        except Exception:  # pylint: disable=broad-except
            trace = traceback.format_exc()
            log.error(
                "Function %s in mine.update failed to execute",
                function_name or function_alias,
            )
            log.debug("Error: %s", trace)
            continue
        if minion_acl.get("allow_tgt"):
            mine_data[function_alias] = salt.utils.mine.wrap_acl_structure(
                res, **minion_acl
            )
        else:
            mine_data[function_alias] = res
    return _mine_store(mine_data, clear)


def send(name, *args, **kwargs):
    """
    Send a specific function and its result to the salt mine.
    This gets stored in either the local cache, or the salt master's cache.

    :param str name: Name of the function to add to the mine.

    The following pameters are extracted from kwargs if present:

    :param str mine_function: The name of the execution_module.function to run
        and whose value will be stored in the salt mine. Defaults to ``name``.
    :param str allow_tgt: Targeting specification for ACL. Specifies which minions
        are allowed to access this function. Please note both your master and
        minion need to be on, at least, version 3000 for this to work properly.

    :param str allow_tgt_type: Type of the targeting specification. This value will
        be ignored if ``allow_tgt`` is not specified. Please note both your
        master and minion need to be on, at least, version 3000 for this to work
        properly.

    Remaining args and kwargs will be passed on to the function to run.

    :rtype: bool
    :return: Whether executing the function and storing the information was successful.

    .. versionchanged:: 3000

        Added ``allow_tgt``- and ``allow_tgt_type``-parameters to specify which
        minions are allowed to access this function.
        See :ref:`targeting` for more information about targeting.

    CLI Example:

    .. code-block:: bash

        salt '*' mine.send network.ip_addrs interface=eth0
        salt '*' mine.send eth0_ip_addrs mine_function=network.ip_addrs interface=eth0
        salt '*' mine.send eth0_ip_addrs mine_function=network.ip_addrs interface=eth0 allow_tgt='G@grain:value' allow_tgt_type=compound
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    mine_function = kwargs.pop("mine_function", None)
    allow_tgt = kwargs.pop("allow_tgt", None)
    allow_tgt_type = kwargs.pop("allow_tgt_type", None)
    mine_data = {}
    try:
        res = salt.utils.functools.call_function(
            __salt__[mine_function or name], *args, **kwargs
        )
    except Exception as exc:  # pylint: disable=broad-except
        trace = traceback.format_exc()
        log.error("Function %s in mine.send failed to execute", mine_function or name)
        log.debug("Error: %s", trace)
        return False

    if allow_tgt:
        mine_data[name] = salt.utils.mine.wrap_acl_structure(
            res, allow_tgt=allow_tgt, allow_tgt_type=allow_tgt_type
        )
    else:
        mine_data[name] = res
    return _mine_store(mine_data)


def get(tgt, fun, tgt_type="glob", exclude_minion=False):
    """
    Get data from the mine.

    :param str tgt: Target whose mine data to get.
    :param fun: Function to get the mine data of. You can specify multiple functions
        to retrieve using either a list or a comma-separated string of functions.
    :type fun: str or list
    :param str tgt_type: Default ``glob``. Target type to use with ``tgt``.
        See :ref:`targeting` for more information.
        Note that all pillar matches, whether using the compound matching system or
        the pillar matching system, will be exact matches, with globbing disabled.
    :param bool exclude_minion: Excludes the current minion from the result set.

    CLI Example:

    .. code-block:: bash

        salt '*' mine.get '*' network.interfaces
        salt '*' mine.get 'os:Fedora' network.interfaces grain
        salt '*' mine.get 'G@os:Fedora and S@192.168.5.0/24' network.ipaddrs compound

    .. seealso:: Retrieving Mine data from Pillar and Orchestrate

        This execution module is intended to be executed on minions.
        Master-side operations such as Pillar or Orchestrate that require Mine
        data should use the :py:mod:`Mine Runner module <salt.runners.mine>`
        instead; it can be invoked from a Pillar SLS file using the
        :py:func:`saltutil.runner <salt.modules.saltutil.runner>` module. For
        example:

        .. code-block:: jinja

            {% set minion_ips = salt.saltutil.runner('mine.get',
                tgt='*',
                fun='network.ip_addrs',
                tgt_type='glob') %}
    """
    # Load from local minion's cache
    if __opts__["file_client"] == "local":
        ret = {}

        is_target = __salt__[f"match.{tgt_type}"](tgt)
        if not is_target:
            return ret

        data = __salt__["data.get"]("mine_cache")
        if not isinstance(data, dict):
            return ret

        if isinstance(fun, str):
            functions = list(set(fun.split(",")))
            _ret_dict = len(functions) > 1
        elif isinstance(fun, list):
            functions = fun
            _ret_dict = True
        else:
            return ret

        for function in functions:
            if function not in data:
                continue
            # If this is a mine item with minion_side_ACL, get its data
            if salt.utils.mine.MINE_ITEM_ACL_ID in data[function]:
                res = data[function][salt.utils.mine.MINE_ITEM_ACL_DATA]
            else:
                # Backwards compatibility with non-ACL mine data.
                res = data[function]
            if _ret_dict:
                ret.setdefault(function, {})[__opts__["id"]] = res
            else:
                ret[__opts__["id"]] = res
        return ret

    # Load from master
    load = {
        "cmd": "_mine_get",
        "id": __opts__["id"],
        "tgt": tgt,
        "fun": fun,
        "tgt_type": tgt_type,
    }
    ret = _mine_get(load, __opts__)
    if exclude_minion and __opts__["id"] in ret:
        del ret[__opts__["id"]]
    return ret


def delete(fun):
    """
    Remove specific function contents of minion.

    :param str fun: The name of the function.
    :rtype: bool
    :return: True on success.

    CLI Example:

    .. code-block:: bash

        salt '*' mine.delete 'network.interfaces'
    """
    if __opts__["file_client"] == "local":
        data = __salt__["data.get"]("mine_cache")
        if isinstance(data, dict) and fun in data:
            del data[fun]
        return __salt__["data.update"]("mine_cache", data)
    load = {
        "cmd": "_mine_delete",
        "id": __opts__["id"],
        "fun": fun,
    }
    return _mine_send(load, __opts__)


def flush():
    """
    Remove all mine contents of minion.

    :rtype: bool
    :return: True on success

    CLI Example:

    .. code-block:: bash

        salt '*' mine.flush
    """
    if __opts__["file_client"] == "local":
        return __salt__["data.update"]("mine_cache", {})
    load = {
        "cmd": "_mine_flush",
        "id": __opts__["id"],
    }
    return _mine_send(load, __opts__)


def get_docker(interfaces=None, cidrs=None, with_container_id=False):
    """
    .. versionchanged:: 2017.7.8,2018.3.3
        When :conf_minion:`docker.update_mine` is set to ``False`` for a given
        minion, no mine data will be populated for that minion, and thus none
        will be returned for it.
    .. versionchanged:: 2019.2.0
        :conf_minion:`docker.update_mine` now defaults to ``False``

    Get all mine data for :py:func:`docker.ps <salt.modules.dockermod.ps_>` and
    run an aggregation routine. The ``interfaces`` parameter allows for
    specifying the network interfaces from which to select IP addresses. The
    ``cidrs`` parameter allows for specifying a list of subnets which the IP
    address must match.

    with_container_id
        Boolean, to expose container_id in the list of results

        .. versionadded:: 2015.8.2

    CLI Example:

    .. code-block:: bash

        salt '*' mine.get_docker
        salt '*' mine.get_docker interfaces='eth0'
        salt '*' mine.get_docker interfaces='["eth0", "eth1"]'
        salt '*' mine.get_docker cidrs='107.170.147.0/24'
        salt '*' mine.get_docker cidrs='["107.170.147.0/24", "172.17.42.0/24"]'
        salt '*' mine.get_docker interfaces='["eth0", "eth1"]' cidrs='["107.170.147.0/24", "172.17.42.0/24"]'
    """
    # Enforce that interface and cidr are lists
    if interfaces:
        interface_ = []
        interface_.extend(interfaces if isinstance(interfaces, list) else [interfaces])
        interfaces = interface_
    if cidrs:
        cidr_ = []
        cidr_.extend(cidrs if isinstance(cidrs, list) else [cidrs])
        cidrs = cidr_

    # Get docker info
    cmd = "docker.ps"
    docker_hosts = get("*", cmd)

    proxy_lists = {}

    # Process docker info
    for containers in docker_hosts.values():
        host = containers.pop("host")
        host_ips = []

        # Prepare host_ips list
        if not interfaces:
            for info in host["interfaces"].values():
                if "inet" in info:
                    for ip_ in info["inet"]:
                        host_ips.append(ip_["address"])
        else:
            for interface in interfaces:
                if interface in host["interfaces"]:
                    if "inet" in host["interfaces"][interface]:
                        for item in host["interfaces"][interface]["inet"]:
                            host_ips.append(item["address"])
        host_ips = list(set(host_ips))

        # Filter out ips from host_ips with cidrs
        if cidrs:
            good_ips = []
            for cidr in cidrs:
                for ip_ in host_ips:
                    if salt.utils.network.in_subnet(cidr, [ip_]):
                        good_ips.append(ip_)
            host_ips = list(set(good_ips))

        # Process each container
        for container in containers.values():
            container_id = container["Info"]["Id"]
            if container["Image"] not in proxy_lists:
                proxy_lists[container["Image"]] = {}
            for dock_port in container["Ports"]:
                # IP exists only if port is exposed
                ip_address = dock_port.get("IP")
                # If port is 0.0.0.0, then we must get the docker host IP
                if ip_address == "0.0.0.0":
                    for ip_ in host_ips:
                        containers = (
                            proxy_lists[container["Image"]]
                            .setdefault("ipv4", {})
                            .setdefault(dock_port["PrivatePort"], [])
                        )
                        container_network_footprint = "{}:{}".format(
                            ip_, dock_port["PublicPort"]
                        )
                        if with_container_id:
                            value = (container_network_footprint, container_id)
                        else:
                            value = container_network_footprint
                        if value not in containers:
                            containers.append(value)
                elif ip_address:
                    containers = (
                        proxy_lists[container["Image"]]
                        .setdefault("ipv4", {})
                        .setdefault(dock_port["PrivatePort"], [])
                    )
                    container_network_footprint = "{}:{}".format(
                        dock_port["IP"], dock_port["PublicPort"]
                    )
                    if with_container_id:
                        value = (container_network_footprint, container_id)
                    else:
                        value = container_network_footprint
                    if value not in containers:
                        containers.append(value)

    return proxy_lists


def valid():
    """
    List valid entries in mine configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' mine.valid
    """
    mine_functions = __salt__["config.merge"]("mine_functions", {})
    # If we don't have any mine functions configured, then we should just bail out
    if not mine_functions:
        return

    mine_data = {}
    for function_alias, function_data in mine_functions.items():
        (
            function_name,
            function_args,
            function_kwargs,
            minion_acl,
        ) = salt.utils.mine.parse_function_definition(function_data)
        if not _mine_function_available(function_name or function_alias):
            continue
        if function_name:
            mine_data[function_alias] = {
                function_name: function_args
                + [{key: value} for key, value in function_kwargs.items()]
            }
        else:
            mine_data[function_alias] = function_data
    return mine_data



import pytest

import salt.auth.pam
from tests.support.mock import patch

pytestmark = [
    pytest.mark.skip_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {salt.auth.pam: {}}


@pytest.fixture
def mock_pam():
    with patch("salt.auth.pam.CALLOC", autospec=True), patch(
        "salt.auth.pam.pointer", autospec=True
    ), patch("salt.auth.pam.PamHandle", autospec=True), patch(
        "salt.auth.pam.PAM_START", autospec=True, return_value=0
    ), patch(
        "salt.auth.pam.PAM_AUTHENTICATE", autospec=True, return_value=0
    ), patch(
        "salt.auth.pam.PAM_END", autospec=True
    ):
        yield


def test_cve_if_pam_acct_mgmt_returns_nonzero_authenticate_should_be_false(mock_pam):
    with patch("salt.auth.pam.PAM_ACCT_MGMT", autospec=True, return_value=42):
        assert (
            salt.auth.pam._authenticate(
                username="fnord", password="fnord", service="login", encoding="utf-8"
            )
            is False
        )


def test_if_pam_acct_mgmt_returns_zero_authenticate_should_be_true(mock_pam):
    with patch("salt.auth.pam.PAM_ACCT_MGMT", autospec=True, return_value=0):
        assert (
            salt.auth.pam._authenticate(
                username="fnord", password="fnord", service="login", encoding="utf-8"
            )
            is True
        )



"""
    Unit tests for salt.utils.yamlloader.SaltYamlSafeLoader
"""

import textwrap

from yaml.constructor import ConstructorError

import salt.utils.files
from salt.utils.yamlloader import SaltYamlSafeLoader, yaml
from tests.support.mock import mock_open, patch
from tests.support.unit import TestCase


class YamlLoaderTestCase(TestCase):
    """
    TestCase for salt.utils.yamlloader module
    """

    @staticmethod
    def render_yaml(data):
        """
        Takes a YAML string, puts it into a mock file, passes that to the YAML
        SaltYamlSafeLoader and then returns the rendered/parsed YAML data
        """
        with patch("salt.utils.files.fopen", mock_open(read_data=data)) as mocked_file:
            with salt.utils.files.fopen(mocked_file) as mocked_stream:
                return SaltYamlSafeLoader(mocked_stream).get_data()

    def test_yaml_basics(self):
        """
        Test parsing an ordinary path
        """
        self.assertEqual(
            self.render_yaml(
                textwrap.dedent(
                    """\
                p1:
                  - alpha
                  - beta"""
                )
            ),
            {"p1": ["alpha", "beta"]},
        )

    def test_yaml_merge(self):
        """
        Test YAML anchors
        """
        # Simple merge test
        self.assertEqual(
            self.render_yaml(
                textwrap.dedent(
                    """\
                p1: &p1
                  v1: alpha
                p2:
                  <<: *p1
                  v2: beta"""
                )
            ),
            {"p1": {"v1": "alpha"}, "p2": {"v1": "alpha", "v2": "beta"}},
        )

        # Test that keys/nodes are overwritten
        self.assertEqual(
            self.render_yaml(
                textwrap.dedent(
                    """\
                p1: &p1
                  v1: alpha
                p2:
                  <<: *p1
                  v1: new_alpha"""
                )
            ),
            {"p1": {"v1": "alpha"}, "p2": {"v1": "new_alpha"}},
        )

        # Test merging of lists
        self.assertEqual(
            self.render_yaml(
                textwrap.dedent(
                    """\
                p1: &p1
                  v1: &v1
                    - t1
                    - t2
                p2:
                  v2: *v1"""
                )
            ),
            {"p2": {"v2": ["t1", "t2"]}, "p1": {"v1": ["t1", "t2"]}},
        )

    def test_yaml_duplicates(self):
        """
        Test that duplicates still throw an error
        """
        with self.assertRaises(ConstructorError):
            self.render_yaml(
                textwrap.dedent(
                    """\
                p1: alpha
                p1: beta"""
                )
            )

        with self.assertRaises(ConstructorError):
            self.render_yaml(
                textwrap.dedent(
                    """\
                p1: &p1
                  v1: alpha
                p2:
                  <<: *p1
                  v2: beta
                  v2: betabeta"""
                )
            )

    def test_yaml_with_plain_scalars(self):
        """
        Test that plain (i.e. unqoted) string and non-string scalars are
        properly handled
        """
        self.assertEqual(
            self.render_yaml(
                textwrap.dedent(
                    """\
                foo:
                  b: {foo: bar, one: 1, list: [1, two, 3]}"""
                )
            ),
            {"foo": {"b": {"foo": "bar", "one": 1, "list": [1, "two", 3]}}},
        )

    def test_not_yaml_monkey_patching(self):
        if hasattr(yaml, "CSafeLoader"):
            assert yaml.SafeLoader != yaml.CSafeLoader



"""
Unit tests for salt.utils.compat.py
"""

import pytest

import salt.utils.compat


def test_cmp_int_x_equals_y():
    # int x == int y
    ret = salt.utils.compat.cmp(1, 1)
    assert ret == 0


def test_cmp_int_x_less_than_y():
    # int x < int y
    ret = salt.utils.compat.cmp(1, 2)
    assert ret == -1


def test_cmp_int_x_greater_than_y():
    # int x > int y
    ret = salt.utils.compat.cmp(2, 1)
    assert ret == 1


def test_cmp_dict_x_equals_y():
    # dict x == dict y
    dict1 = {"foo": "bar", "baz": "qux"}
    dict2 = {"baz": "qux", "foo": "bar"}
    ret = salt.utils.compat.cmp(dict1, dict2)
    assert ret == 0


def test_cmp_dict_x_not_equals_y():
    # dict x != dict y
    dict1 = {"foo": "bar", "baz": "qux"}
    dict2 = {"foobar": "bar", "baz": "qux"}
    ret = salt.utils.compat.cmp(dict1, dict2)
    assert ret == -1


def test_cmp_dict_x_not_equals_int_y():
    # dict x != int y
    dict1 = {"foo": "bar", "baz": "qux"}
    pytest.raises(TypeError, salt.utils.compat.cmp, dict1, 1)



"""
Platform independent versions of some os/os.path functions. Gets around PY2's
lack of support for reading NTFS links.
"""

import logging
import os
import posixpath
import re
from collections.abc import Iterable

import salt.utils.args
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandNotFoundError
from salt.utils.decorators.jinja import jinja_filter

try:
    import win32con
    import win32file

    HAS_WIN32FILE = True
except ImportError:
    HAS_WIN32FILE = False

log = logging.getLogger(__name__)


def islink(path):
    """
    Equivalent to os.path.islink().

    Returns True for Symlinks and Junctions on Windows. Junctions are symlinks
    that only apply to directories. They are specific to Windows. Junctions are
    not handled by `os.path.islink` until Python 3.12.
    """
    # On Windows we want to detect reparse points to detect both symlinks and
    # junctions as links until os.path.islink handles junctions.
    if salt.utils.platform.is_windows():
        return _is_reparse_point(path)

    return os.path.islink(path)


def readlink(path):
    """
    Equivalent to os.readlink()
    """
    base = os.readlink(path)
    if salt.utils.platform.is_windows():
        # Python 3.8 added support for directory junctions which prefixes the
        # return with `\\?\`. We need to strip that off.
        # https://docs.python.org/3/library/os.html#os.readlink
        if base.startswith("\\\\?\\"):
            base = base[4:]
    return base


def _is_reparse_point(path):
    """
    A Symbolic Link and a Junction are both reparse points on Windows.
    Returns True if path is a reparse point; False otherwise.
    """
    result = win32file.GetFileAttributesW(path)

    # An error has occurred, so this is not a symlink
    if result == -1:
        return False

    # If the Reparse Point bit is set, then it is a reparse point
    return True if result & win32con.FILE_ATTRIBUTE_REPARSE_POINT else False


def _get_reparse_data(path):
    """
    Retrieves the reparse point data structure for the given path.

    If the path is not a reparse point, None is returned.

    See http://msdn.microsoft.com/en-us/library/ff552012.aspx for details on the
    REPARSE_DATA_BUFFER structure returned.
    """
    # ensure paths are using the right slashes
    path = os.path.normpath(path)

    if not _is_reparse_point(path):
        return None

    fileHandle = None
    try:
        fileHandle = win32file.CreateFileW(
            path,
            0x80000000,  # GENERIC_READ
            1,  # share with other readers
            None,  # no inherit, default security descriptor
            3,  # OPEN_EXISTING
            0x00200000
            | 0x02000000,  # FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS
        )

        reparseData = win32file.DeviceIoControl(
            fileHandle,
            0x900A8,  # FSCTL_GET_REPARSE_POINT
            None,  # in buffer
            16384,  # out buffer size (MAXIMUM_REPARSE_DATA_BUFFER_SIZE)
        )

    finally:
        if fileHandle:
            win32file.CloseHandle(fileHandle)

    return reparseData


@jinja_filter("which")
def which(exe=None):
    """
    Python clone of /usr/bin/which
    """

    if not exe:
        log.error("No executable was passed to be searched by salt.utils.path.which()")
        return None

    ## define some utilities (we use closures here because our predecessor used them)
    def is_executable_common(path):
        """
        This returns truth if posixy semantics (which python simulates on
        windows) states that this is executable.
        """
        return os.path.isfile(path) and os.access(path, os.X_OK)

    def resolve(path):
        """
        This will take a path and recursively follow the link until we get to a
        real file.
        """
        while os.path.islink(path):
            res = readlink(path)

            # if the link points to a relative target, then convert it to an
            # absolute path relative to the original path
            if not os.path.isabs(res):
                directory, _ = os.path.split(path)
                res = join(directory, res)
            path = res
        return path

    # windows-only
    def has_executable_ext(path, ext_membership):
        """
        Extract the extension from the specified path, lowercase it so we
        can be insensitive, and then check it against the available exts.
        """
        p, ext = os.path.splitext(path)
        return ext.lower() in ext_membership

    ## prepare related variables from the environment
    res = salt.utils.stringutils.to_unicode(os.environ.get("PATH", ""))
    system_path = res.split(os.pathsep)

    # add some reasonable defaults in case someone's PATH is busted
    if not salt.utils.platform.is_windows():
        res = set(system_path)
        extended_path = [
            "/sbin",
            "/bin",
            "/usr/sbin",
            "/usr/bin",
            "/usr/local/sbin",
            "/usr/local/bin",
        ]
        system_path.extend([p for p in extended_path if p not in res])

    ## now to define the semantics of what's considered executable on a given platform
    if salt.utils.platform.is_windows():
        # executable semantics on windows requires us to search PATHEXT
        res = salt.utils.stringutils.to_str(os.environ.get("PATHEXT", ".EXE"))

        # generate two variables, one of them for O(n) searches (but ordered)
        # and another for O(1) searches. the previous guy was trying to use
        # memoization with a function that has no arguments, this provides
        # the exact same benefit
        pathext = res.split(os.pathsep)
        res = {ext.lower() for ext in pathext}

        # check if our caller already specified a valid extension as then we don't need to match it
        _, ext = os.path.splitext(exe)
        if ext.lower() in res:
            pathext = [""]

            is_executable = is_executable_common

        # The specified extension isn't valid, so we just assume it's part of the
        # filename and proceed to walk the pathext list
        else:

            def is_executable(path, membership=res):
                return is_executable_common(path) and has_executable_ext(
                    path, membership
                )

    else:
        # in posix, there's no such thing as file extensions..only zuul
        pathext = [""]

        # executable semantics are pretty simple on reasonable platforms...
        is_executable = is_executable_common

    ## search for the executable

    # check to see if the full path was specified as then we don't need
    # to actually walk the system_path for any reason
    if is_executable(exe):
        return exe

    # now to search through our system_path
    for path in system_path:
        p = join(os.path.expandvars(path), exe)

        # iterate through all extensions to see which one is executable
        for ext in pathext:
            pext = p + ext
            rp = resolve(pext)
            if is_executable(rp):
                return p + ext
            continue
        continue

    ## if something was executable, we should've found it already...
    log.trace(
        "'%s' could not be found in the following search path: '%s'", exe, system_path
    )
    return None


def which_bin(exes):
    """
    Scan over some possible executables and return the first one that is found
    """
    if isinstance(exes, str):
        exes = [exes]
    if not isinstance(exes, Iterable):
        return None
    for exe in exes:
        path = which(exe)
        if not path:
            continue
        return path
    return None


@jinja_filter("path_join")
def join(*parts, **kwargs):
    """
    This functions tries to solve some issues when joining multiple absolute
    paths on both *nix and windows platforms.

    See tests/unit/utils/path_join_test.py for some examples on what's being
    talked about here.

    The "use_posixpath" kwarg can be be used to force joining using poxixpath,
    which is useful for Salt fileserver paths on Windows masters.
    """
    parts = [salt.utils.stringutils.to_str(part) for part in parts]

    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    use_posixpath = kwargs.pop("use_posixpath", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    pathlib = posixpath if use_posixpath else os.path

    # Normalize path converting any os.sep as needed
    parts = [pathlib.normpath(p) for p in parts]

    try:
        root = parts.pop(0)
    except IndexError:
        # No args passed to func
        return ""

    root = salt.utils.stringutils.to_unicode(root)
    if not parts:
        ret = root
    else:
        stripped = [p.lstrip(os.sep) for p in parts]
        ret = pathlib.join(root, *salt.utils.data.decode(stripped))
    return pathlib.normpath(ret)


def check_or_die(command):
    """
    Simple convenience function for modules to use for gracefully blowing up
    if a required tool is not available in the system path.

    Lazily import `salt.modules.cmdmod` to avoid any sort of circular
    dependencies.
    """
    if command is None:
        raise CommandNotFoundError("'None' is not a valid command.")

    if not which(command):
        raise CommandNotFoundError(f"'{command}' is not in the path")


def sanitize_win_path(winpath):
    """
    Remove illegal path characters for windows
    """
    intab = "<>:|?*"
    if isinstance(winpath, str):
        winpath = winpath.translate({ord(c): "_" for c in intab})
    elif isinstance(winpath, str):
        outtab = "_" * len(intab)
        trantab = "".maketrans(intab, outtab)
        winpath = winpath.translate(trantab)
    return winpath


def safe_path(path, allow_path=None):
    r"""
    .. versionadded:: 2017.7.3

    Checks that the path is safe for modification by Salt. For example, you
    wouldn't want to have salt delete the contents of ``C:\Windows``. The
    following directories are considered unsafe:

    - C:\, D:\, E:\, etc.
    - \
    - C:\Windows

    Args:

        path (str): The path to check

        allow_paths (str, list): A directory or list of directories inside of
            path that may be safe. For example: ``C:\Windows\TEMP``

    Returns:
        bool: True if safe, otherwise False
    """
    # Create regex definitions for directories that may be unsafe to modify
    system_root = os.environ.get("SystemRoot", "C:\\Windows")
    deny_paths = (
        r"[a-z]\:\\$",  # C:\, D:\, etc
        r"\\$",  # \
        re.escape(system_root),  # C:\Windows
    )

    # Make allow_path a list
    if allow_path and not isinstance(allow_path, list):
        allow_path = [allow_path]

    # Create regex definition for directories we may want to make exceptions for
    allow_paths = list()
    if allow_path:
        for item in allow_path:
            allow_paths.append(re.escape(item))

    # Check the path to make sure it's not one of the bad paths
    good_path = True
    for d_path in deny_paths:
        if re.match(d_path, path, flags=re.IGNORECASE) is not None:
            # Found deny path
            good_path = False

    # If local_dest is one of the bad paths, check for exceptions
    if not good_path:
        for a_path in allow_paths:
            if re.match(a_path, path, flags=re.IGNORECASE) is not None:
                # Found exception
                good_path = True

    return good_path


def os_walk(top, *args, **kwargs):
    """
    This is a helper than ensures that all paths returned from os.walk are
    unicode.
    """
    top_query = salt.utils.stringutils.to_str(top)
    for item in os.walk(top_query, *args, **kwargs):
        yield salt.utils.data.decode(item, preserve_tuples=True)


def expand(path):
    """
    Expands all user and environment variables
    .. versionadded:: 3005

    Args:

        path (str): A path to a file or directory

    Returns:
        str: A fully expanded, real path
    """
    path = os.path.expanduser(path)  # expand ~ to home directory
    path = os.path.expandvars(path)  # expand any other environment vars
    return os.path.realpath(path)  # fix path format



"""
tests.pytests.integration.reactor.test_reactor
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Test Salt's reactor system
"""

import logging
import pathlib
import time
import types

import pytest

import salt.utils.event
import salt.utils.reactor
from salt.serializers import yaml
from tests.support.helpers import PRE_PYTEST_SKIP_REASON

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON),
]

log = logging.getLogger(__name__)


@pytest.fixture
def master_event_bus(salt_master):
    with salt.utils.event.get_event(
        "master",
        opts=salt_master.config.copy(),
        sock_dir=salt_master.config["sock_dir"],
        listen=True,
        raise_errors=True,
    ) as event:
        yield event


@pytest.fixture
def minion_event_bus(salt_minion):
    with salt.utils.event.get_event(
        "minion",
        opts=salt_minion.config.copy(),
        sock_dir=salt_minion.config["sock_dir"],
        listen=True,
        raise_errors=True,
    ) as event:
        yield event


@pytest.fixture
def event_listerner_timeout(grains):
    if grains["os"] == "Windows":
        if grains["osrelease"].startswith("2019"):
            return types.SimpleNamespace(catch=120, miss=30)
        return types.SimpleNamespace(catch=90, miss=10)
    return types.SimpleNamespace(catch=60, miss=10)


def test_ping_reaction(
    event_listener, salt_minion, minion_event_bus, event_listerner_timeout
):
    """
    Fire an event on the master and ensure that it pings the minion
    """
    event_tag = "reactor/test-ping-reaction"
    start_time = time.time()
    # Send test event
    minion_event_bus.fire_event({"a": "b"}, event_tag)

    event_pattern = (salt_minion.id, event_tag)
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=event_listerner_timeout.catch
    )
    assert matched_events.found_all_events
    for event in matched_events:
        assert event.data == {"a": "b"}


@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
def test_reactor_reaction(
    event_listener,
    salt_master,
    salt_minion,
    master_event_bus,
    reactor_event,
    event_listerner_timeout,
):
    """
    Fire an event on the master and ensure the reactor event responds
    """

    start_time = time.time()
    master_event_bus.fire_event({"id": salt_minion.id}, reactor_event.tag)
    event_pattern = (salt_master.id, reactor_event.event_tag)
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=event_listerner_timeout.catch
    )
    assert matched_events.found_all_events
    for event in matched_events:
        assert event.data["test_reaction"] is True


@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.timeout_unless_on_windows(120)
def test_reactor_is_leader(
    event_listener,
    salt_master,
    salt_run_cli,
    master_event_bus,
    reactor_event,
    salt_minion,
    event_listerner_timeout,
):
    """
    If reactor system is unavailable, an exception is thrown.
    When leader is true (the default), the reacion event should return.
    When leader is set to false reactor should timeout/not do anything.
    """
    ret = salt_run_cli.run("reactor.is_leader")
    assert ret.returncode == 1
    assert (
        "salt.exceptions.CommandExecutionError: Reactor system is not running."
        in ret.stdout
    )

    ret = salt_run_cli.run("reactor.set_leader", value=True)
    assert ret.returncode == 1
    assert (
        "salt.exceptions.CommandExecutionError: Reactor system is not running."
        in ret.stdout
    )

    ret = salt_run_cli.run("reactor.is_leader")
    assert ret.returncode == 1
    assert (
        "salt.exceptions.CommandExecutionError: Reactor system is not running."
        in ret.stdout
    )

    # make reactor not the leader; ensure reactor engine is available
    engines_config = salt_master.config.get("engines").copy()
    for idx, engine in enumerate(list(engines_config)):
        if "reactor" in engine:
            engines_config.pop(idx)

    engines_config.append(
        {
            "reactor": {
                "refresh_interval": 60,
                "worker_threads": 10,
                "worker_hwm": 10000,
            }
        }
    )
    config_overrides = yaml.serialize({"engines": engines_config})
    confd_dir = (
        pathlib.Path(salt_master.config_dir)
        / pathlib.Path(salt_master.config["default_include"]).parent
    )
    confd_dir.mkdir(exist_ok=True)

    # Now, with the temp config in place, ensure the reactor engine is running
    with pytest.helpers.temp_file("reactor-test.conf", config_overrides, confd_dir):
        ret = salt_run_cli.run("reactor.set_leader", value=True)
        assert ret.returncode == 0
        assert (
            "CommandExecutionError" not in ret.stdout
        ), "reactor engine is not running"

        ret = salt_run_cli.run("reactor.is_leader")
        assert ret.returncode == 0
        assert ret.stdout.rstrip().splitlines()[-1] == "true"

        ret = salt_run_cli.run("reactor.set_leader", value=False)
        assert ret.returncode == 0

        ret = salt_run_cli.run("reactor.is_leader")
        assert ret.returncode == 0
        assert ret.stdout.rstrip().splitlines()[-1] == "false"

        start_time = time.time()
        master_event_bus.fire_event({"id": salt_minion.id}, reactor_event.tag)

        # Since leader is false, let's try and get the fire event to ensure it was triggered
        event_pattern = (salt_master.id, reactor_event.tag)
        matched_events = event_listener.wait_for_events(
            [event_pattern],
            after_time=start_time,
            timeout=event_listerner_timeout.catch,
        )
        assert matched_events.found_all_events
        # Now that we matched the trigger event, let's confirm we don't get the reaction event
        event_pattern = (salt_master.id, reactor_event.event_tag)
        matched_events = event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=event_listerner_timeout.miss
        )
        assert matched_events.found_all_events is not True

        # make reactor the leader again; ensure reactor engine is available
        ret = salt_run_cli.run("reactor.set_leader", value=True)
        assert ret.returncode == 0
        ret = salt_run_cli.run("reactor.is_leader")
        assert ret.returncode == 0
        assert ret.stdout.rstrip().splitlines()[-1] == "true"

        # trigger a reaction
        start_time = time.time()
        master_event_bus.fire_event({"id": salt_minion.id}, reactor_event.tag)
        event_pattern = (salt_master.id, reactor_event.event_tag)
        matched_events = event_listener.wait_for_events(
            [event_pattern],
            after_time=start_time,
            timeout=event_listerner_timeout.catch,
        )
        assert matched_events.found_all_events
        for event in matched_events:
            assert event.data["test_reaction"] is True

    # Let's just confirm the engine is not running once again(because the config file is deleted by now)
    ret = salt_run_cli.run("reactor.is_leader")
    assert ret.returncode == 1
    assert (
        "salt.exceptions.CommandExecutionError: Reactor system is not running."
        in ret.stdout
    )



import pytest
from saltfactories.utils import random_string

from tests.conftest import FIPS_TESTRUN


@pytest.fixture(scope="package")
def salt_master_factory(salt_factories):
    factory = salt_factories.salt_master_daemon(
        random_string("reauth-master-"),
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "publish_signing_algorithm": (
                "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
            ),
        },
    )
    return factory


@pytest.fixture(scope="package")
def salt_master(salt_master_factory):
    with salt_master_factory.started():
        yield salt_master_factory


@pytest.fixture(scope="package")
def salt_minion_factory(salt_master):
    factory = salt_master.salt_minion_daemon(
        random_string("reauth-minion-"),
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    return factory


@pytest.fixture(scope="package")
def salt_minion(salt_minion_factory):
    with salt_minion_factory.started():
        yield salt_minion_factory


@pytest.fixture(scope="package")
def salt_key_cli(salt_master):
    assert salt_master.is_running()
    return salt_master.salt_key_cli()


@pytest.fixture(scope="package")
def salt_cli(salt_master):
    assert salt_master.is_running()
    return salt_master.salt_cli()



import os
import shutil
import stat
import sys
import tempfile

import pytest

import salt.utils.files
import salt.utils.find
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class TestFind(TestCase):
    def test_parse_interval(self):
        self.assertRaises(ValueError, salt.utils.find._parse_interval, "w")
        self.assertRaises(ValueError, salt.utils.find._parse_interval, "1")
        self.assertRaises(ValueError, salt.utils.find._parse_interval, "1s1w")
        self.assertRaises(ValueError, salt.utils.find._parse_interval, "1s1s")

        result, resolution, modifier = salt.utils.find._parse_interval("")
        self.assertEqual(result, 0)
        self.assertIs(resolution, None)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1s")
        self.assertEqual(result, 1.0)
        self.assertEqual(resolution, 1)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1m")
        self.assertEqual(result, 60.0)
        self.assertEqual(resolution, 60)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1h")
        self.assertEqual(result, 3600.0)
        self.assertEqual(resolution, 3600)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1d")
        self.assertEqual(result, 86400.0)
        self.assertEqual(resolution, 86400)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1w")
        self.assertEqual(result, 604800.0)
        self.assertEqual(resolution, 604800)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1w3d6h")
        self.assertEqual(result, 885600.0)
        self.assertEqual(resolution, 3600)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1m1s")
        self.assertEqual(result, 61.0)
        self.assertEqual(resolution, 1)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("1m2s")
        self.assertEqual(result, 62.0)
        self.assertEqual(resolution, 1)
        self.assertEqual(modifier, "")

        result, resolution, modifier = salt.utils.find._parse_interval("+1d")
        self.assertEqual(result, 86400.0)
        self.assertEqual(resolution, 86400)
        self.assertEqual(modifier, "+")

        result, resolution, modifier = salt.utils.find._parse_interval("-1d")
        self.assertEqual(result, 86400.0)
        self.assertEqual(resolution, 86400)
        self.assertEqual(modifier, "-")

    def test_parse_size(self):
        self.assertRaises(ValueError, salt.utils.find._parse_size, "")
        self.assertRaises(ValueError, salt.utils.find._parse_size, "1s1s")
        min_size, max_size = salt.utils.find._parse_size("1")
        self.assertEqual(min_size, 1)
        self.assertEqual(max_size, 1)

        min_size, max_size = salt.utils.find._parse_size("1b")
        self.assertEqual(min_size, 1)
        self.assertEqual(max_size, 1)

        min_size, max_size = salt.utils.find._parse_size("1k")
        self.assertEqual(min_size, 1024)
        self.assertEqual(max_size, 2047)

        min_size, max_size = salt.utils.find._parse_size("1m")
        self.assertEqual(min_size, 1048576)
        self.assertEqual(max_size, 2097151)

        min_size, max_size = salt.utils.find._parse_size("1g")
        self.assertEqual(min_size, 1073741824)
        self.assertEqual(max_size, 2147483647)

        min_size, max_size = salt.utils.find._parse_size("1t")
        self.assertEqual(min_size, 1099511627776)
        self.assertEqual(max_size, 2199023255551)

        min_size, max_size = salt.utils.find._parse_size("0m")
        self.assertEqual(min_size, 0)
        self.assertEqual(max_size, 1048575)

        min_size, max_size = salt.utils.find._parse_size("-1m")
        self.assertEqual(min_size, 0)
        self.assertEqual(max_size, 1048576)

        min_size, max_size = salt.utils.find._parse_size("+1m")
        self.assertEqual(min_size, 1048576)
        self.assertEqual(max_size, sys.maxsize)

        min_size, max_size = salt.utils.find._parse_size("+1M")
        self.assertEqual(min_size, 1048576)
        self.assertEqual(max_size, sys.maxsize)

    def test_option_requires(self):
        option = salt.utils.find.Option()
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_PATH)

    def test_name_option_match(self):
        option = salt.utils.find.NameOption("name", "*.txt")
        self.assertIs(option.match("", "", ""), None)
        self.assertIs(option.match("", "hello.txt", "").group(), "hello.txt")
        self.assertIs(option.match("", "HELLO.TXT", ""), None)

    def test_iname_option_match(self):
        option = salt.utils.find.InameOption("name", "*.txt")
        self.assertIs(option.match("", "", ""), None)
        self.assertIs(option.match("", "hello.txt", "").group(), "hello.txt")
        self.assertIs(option.match("", "HELLO.TXT", "").group(), "HELLO.TXT")

    def test_regex_option_match(self):
        self.assertRaises(ValueError, salt.utils.find.RegexOption, "name", "(.*}")

        option = salt.utils.find.RegexOption("name", r".*\.txt")
        self.assertIs(option.match("", "", ""), None)
        self.assertIs(option.match("", "hello.txt", "").group(), "hello.txt")
        self.assertIs(option.match("", "HELLO.TXT", ""), None)

    def test_iregex_option_match(self):
        self.assertRaises(ValueError, salt.utils.find.IregexOption, "name", "(.*}")

        option = salt.utils.find.IregexOption("name", r".*\.txt")
        self.assertIs(option.match("", "", ""), None)
        self.assertIs(option.match("", "hello.txt", "").group(), "hello.txt")
        self.assertIs(option.match("", "HELLO.TXT", "").group(), "HELLO.TXT")

    def test_type_option_requires(self):
        self.assertRaises(ValueError, salt.utils.find.TypeOption, "type", "w")

        option = salt.utils.find.TypeOption("type", "d")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

    def test_type_option_match(self):
        option = salt.utils.find.TypeOption("type", "b")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), False)

        option = salt.utils.find.TypeOption("type", "c")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), False)

        option = salt.utils.find.TypeOption("type", "d")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), False)

        option = salt.utils.find.TypeOption("type", "f")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), True)

        option = salt.utils.find.TypeOption("type", "l")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), False)

        option = salt.utils.find.TypeOption("type", "p")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), False)

        option = salt.utils.find.TypeOption("type", "s")
        self.assertEqual(option.match("", "", [stat.S_IFREG]), False)

        option = salt.utils.find.TypeOption("type", "b")
        self.assertEqual(option.match("", "", [stat.S_IFBLK]), True)

        option = salt.utils.find.TypeOption("type", "c")
        self.assertEqual(option.match("", "", [stat.S_IFCHR]), True)

        option = salt.utils.find.TypeOption("type", "d")
        self.assertEqual(option.match("", "", [stat.S_IFDIR]), True)

        option = salt.utils.find.TypeOption("type", "l")
        self.assertEqual(option.match("", "", [stat.S_IFLNK]), True)

        option = salt.utils.find.TypeOption("type", "p")
        self.assertEqual(option.match("", "", [stat.S_IFIFO]), True)

        option = salt.utils.find.TypeOption("type", "s")
        self.assertEqual(option.match("", "", [stat.S_IFSOCK]), True)

    @pytest.mark.skip_on_windows(reason="pwd not available on Windows")
    def test_owner_option_requires(self):
        self.assertRaises(ValueError, salt.utils.find.OwnerOption, "owner", "notexist")

        option = salt.utils.find.OwnerOption("owner", "root")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

    @pytest.mark.skip_on_windows(reason="pwd not available on Windows")
    def test_owner_option_match(self):
        option = salt.utils.find.OwnerOption("owner", "root")
        self.assertEqual(option.match("", "", [0] * 5), True)

        option = salt.utils.find.OwnerOption("owner", "500")
        self.assertEqual(option.match("", "", [500] * 5), True)

    @pytest.mark.skip_on_windows(reason="grp not available on Windows")
    def test_group_option_requires(self):
        self.assertRaises(ValueError, salt.utils.find.GroupOption, "group", "notexist")

        if sys.platform.startswith(("darwin", "freebsd", "openbsd")):
            group_name = "wheel"
        else:
            group_name = "root"
        option = salt.utils.find.GroupOption("group", group_name)
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

    @pytest.mark.skip_on_windows(reason="grp not available on Windows")
    def test_group_option_match(self):
        if sys.platform.startswith(("darwin", "freebsd", "openbsd")):
            group_name = "wheel"
        else:
            group_name = "root"
        option = salt.utils.find.GroupOption("group", group_name)
        self.assertEqual(option.match("", "", [0] * 6), True)

        option = salt.utils.find.GroupOption("group", "500")
        self.assertEqual(option.match("", "", [500] * 6), True)

    def test_size_option_requires(self):
        self.assertRaises(ValueError, salt.utils.find.SizeOption, "size", "1s1s")

        option = salt.utils.find.SizeOption("size", "+1G")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

    def test_size_option_match(self):
        option = salt.utils.find.SizeOption("size", "+1k")
        self.assertEqual(option.match("", "", [10000] * 7), True)

        option = salt.utils.find.SizeOption("size", "+1G")
        self.assertEqual(option.match("", "", [10000] * 7), False)

    def test_mtime_option_requires(self):
        self.assertRaises(ValueError, salt.utils.find.MtimeOption, "mtime", "4g")

        option = salt.utils.find.MtimeOption("mtime", "1d")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

    def test_mtime_option_match(self):
        option = salt.utils.find.MtimeOption("mtime", "-1w")
        self.assertEqual(option.match("", "", [1] * 9), False)

        option = salt.utils.find.MtimeOption("mtime", "-1s")
        self.assertEqual(option.match("", "", [10**10] * 9), True)


class TestGrepOption(TestCase):
    def setUp(self):
        super().setUp()
        self.tmpdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        super().tearDown()

    def test_grep_option_requires(self):
        self.assertRaises(ValueError, salt.utils.find.GrepOption, "grep", "(foo)|(bar}")

        option = salt.utils.find.GrepOption("grep", "(foo)|(bar)")
        find = salt.utils.find
        self.assertEqual(
            option.requires(), (find._REQUIRES_CONTENTS | find._REQUIRES_STAT)
        )

    def test_grep_option_match_regular_file(self):
        hello_file = os.path.join(self.tmpdir, "hello.txt")
        with salt.utils.files.fopen(hello_file, "w") as fp_:
            fp_.write(salt.utils.stringutils.to_str("foo"))
        option = salt.utils.find.GrepOption("grep", "foo")
        self.assertEqual(
            option.match(self.tmpdir, "hello.txt", os.stat(hello_file)), hello_file
        )

        option = salt.utils.find.GrepOption("grep", "bar")
        self.assertEqual(
            option.match(self.tmpdir, "hello.txt", os.stat(hello_file)), None
        )

    @pytest.mark.skip_on_windows(reason="No /dev/null on Windows")
    def test_grep_option_match_dev_null(self):
        option = salt.utils.find.GrepOption("grep", "foo")
        self.assertEqual(option.match("dev", "null", os.stat("/dev/null")), None)


class TestPrintOption(TestCase):
    def setUp(self):
        super().setUp()
        self.tmpdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        super().tearDown()

    def test_print_option_defaults(self):
        option = salt.utils.find.PrintOption("print", "")
        self.assertEqual(option.need_stat, False)
        self.assertEqual(option.print_title, False)
        self.assertEqual(option.fmt, ["path"])

    def test_print_option_requires(self):
        option = salt.utils.find.PrintOption("print", "")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_PATH)

        option = salt.utils.find.PrintOption("print", "name")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_PATH)

        option = salt.utils.find.PrintOption("print", "path")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_PATH)

        option = salt.utils.find.PrintOption("print", "name,path")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_PATH)

        option = salt.utils.find.PrintOption("print", "user")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

        option = salt.utils.find.PrintOption("print", "path user")
        self.assertEqual(option.requires(), salt.utils.find._REQUIRES_STAT)

    @pytest.mark.skip_on_fips_enabled_platform
    def test_print_option_execute(self):
        hello_file = os.path.join(self.tmpdir, "hello.txt")
        with salt.utils.files.fopen(hello_file, "w") as fp_:
            fp_.write(salt.utils.stringutils.to_str("foo"))

        option = salt.utils.find.PrintOption("print", "")
        self.assertEqual(option.execute("", [0] * 9), "")

        option = salt.utils.find.PrintOption("print", "path")
        self.assertEqual(option.execute("test_name", [0] * 9), "test_name")

        option = salt.utils.find.PrintOption("print", "name")
        self.assertEqual(option.execute("test_name", [0] * 9), "test_name")

        option = salt.utils.find.PrintOption("print", "size")
        self.assertEqual(option.execute(hello_file, os.stat(hello_file)), 3)

        option = salt.utils.find.PrintOption("print", "type")
        self.assertEqual(option.execute(hello_file, os.stat(hello_file)), "f")

        option = salt.utils.find.PrintOption("print", "mode")
        self.assertEqual(option.execute(hello_file, range(10)), 0)

        option = salt.utils.find.PrintOption("print", "mtime")
        self.assertEqual(option.execute(hello_file, range(10)), 8)

        option = salt.utils.find.PrintOption("print", "md5")
        self.assertEqual(
            option.execute(hello_file, os.stat(hello_file)),
            "acbd18db4cc2f85cedef654fccc4a4d8",
        )

        option = salt.utils.find.PrintOption("print", "path name")
        self.assertEqual(
            option.execute("test_name", [0] * 9), ["test_name", "test_name"]
        )

        option = salt.utils.find.PrintOption("print", "size name")
        self.assertEqual(option.execute("test_name", [0] * 9), [0, "test_name"])

    @pytest.mark.skip_on_windows(reason="pwd not available on Windows")
    def test_print_user(self):
        option = salt.utils.find.PrintOption("print", "user")
        self.assertEqual(option.execute("", [0] * 10), "root")

        option = salt.utils.find.PrintOption("print", "user")
        self.assertEqual(option.execute("", [2**31] * 10), 2**31)

    @pytest.mark.skip_on_windows(reason="grp not available on Windows")
    def test_print_group(self):
        option = salt.utils.find.PrintOption("print", "group")
        if sys.platform.startswith(("darwin", "freebsd", "openbsd")):
            group_name = "wheel"
        else:
            group_name = "root"
        self.assertEqual(option.execute("", [0] * 10), group_name)

        # This seems to be not working in Ubuntu 12.04 32 bit
        # option = salt.utils.find.PrintOption('print', 'group')
        # self.assertEqual(option.execute('', [2 ** 31] * 10), 2 ** 31)

    @pytest.mark.skip_on_windows(reason="no /dev/null on windows")
    def test_print_md5(self):
        option = salt.utils.find.PrintOption("print", "md5")
        self.assertEqual(option.execute("/dev/null", os.stat("/dev/null")), "")


class TestFinder(TestCase):
    def setUp(self):
        super().setUp()
        self.tmpdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        super().tearDown()

    @pytest.mark.skip_on_windows(reason="No /dev/null on Windows")
    def test_init(self):
        finder = salt.utils.find.Finder({})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(finder.criteria, [])

        finder = salt.utils.find.Finder({"_": None})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(finder.criteria, [])

        self.assertRaises(ValueError, salt.utils.find.Finder, {"": None})
        self.assertRaises(ValueError, salt.utils.find.Finder, {"name": None})
        self.assertRaises(ValueError, salt.utils.find.Finder, {"nonexist": "somevalue"})

        finder = salt.utils.find.Finder({"name": "test_name"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-12:-2], "NameOption")

        finder = salt.utils.find.Finder({"iname": "test_name"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-13:-2], "InameOption")

        finder = salt.utils.find.Finder({"regex": r".*\.txt"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-13:-2], "RegexOption")

        finder = salt.utils.find.Finder({"iregex": r".*\.txt"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-14:-2], "IregexOption")

        finder = salt.utils.find.Finder({"type": "d"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-12:-2], "TypeOption")

        finder = salt.utils.find.Finder({"owner": "root"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-13:-2], "OwnerOption")

        if sys.platform.startswith(("darwin", "freebsd", "openbsd")):
            group_name = "wheel"
        else:
            group_name = "root"
        finder = salt.utils.find.Finder({"group": group_name})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-13:-2], "GroupOption")

        finder = salt.utils.find.Finder({"size": "+1G"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-12:-2], "SizeOption")

        finder = salt.utils.find.Finder({"mtime": "1d"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-13:-2], "MtimeOption")

        finder = salt.utils.find.Finder({"grep": "foo"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(str(finder.criteria[0].__class__)[-12:-2], "GrepOption")

        finder = salt.utils.find.Finder({"print": "name"})
        self.assertEqual(str(finder.actions[0].__class__)[-13:-2], "PrintOption")
        self.assertEqual(finder.criteria, [])

    def test_find(self):
        hello_file = os.path.join(self.tmpdir, "hello.txt")
        with salt.utils.files.fopen(hello_file, "w") as fp_:
            fp_.write(salt.utils.stringutils.to_str("foo"))

        finder = salt.utils.find.Finder({})
        self.assertEqual(list(finder.find(self.tmpdir)), [self.tmpdir, hello_file])

        finder = salt.utils.find.Finder({"mindepth": 1})
        self.assertEqual(list(finder.find(self.tmpdir)), [hello_file])

        finder = salt.utils.find.Finder({"maxdepth": 0})
        self.assertEqual(list(finder.find(self.tmpdir)), [self.tmpdir])

        finder = salt.utils.find.Finder({"name": "hello.txt"})
        self.assertEqual(list(finder.find(self.tmpdir)), [hello_file])

        finder = salt.utils.find.Finder({"type": "f", "print": "path"})
        self.assertEqual(list(finder.find(self.tmpdir)), [hello_file])

        finder = salt.utils.find.Finder({"size": "+1G", "print": "path"})
        self.assertEqual(list(finder.find(self.tmpdir)), [])

        finder = salt.utils.find.Finder({"name": "hello.txt", "print": "path name"})
        self.assertEqual(list(finder.find(self.tmpdir)), [[hello_file, "hello.txt"]])

        finder = salt.utils.find.Finder({"name": "test_name"})
        self.assertEqual(list(finder.find("")), [])



from collections import OrderedDict

import pytest

import salt.states.boto_secgroup as boto_secgroup


@pytest.fixture
def configure_loader_modules():
    return {boto_secgroup: {}}


def test__get_rule_changes_no_rules_no_change():
    """
    tests a condition with no rules in present or desired group
    """
    present_rules = []
    desired_rules = []
    assert boto_secgroup._get_rule_changes(desired_rules, present_rules) == ([], [])


def test__get_rule_changes_create_rules():
    """
    tests a condition where a rule must be created
    """
    present_rules = [
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 22),
                ("to_port", 22),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        )
    ]
    desired_rules = [
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 22),
                ("to_port", 22),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        ),
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 80),
                ("to_port", 80),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        ),
    ]
    # can also use: rules_to_create = [rule for rule in desired_rules if rule not in present_rules]
    rules_to_create = [
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 80),
                ("to_port", 80),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        )
    ]
    assert boto_secgroup._get_rule_changes(desired_rules, present_rules) == (
        [],
        rules_to_create,
    )


def test__get_rule_changes_delete_rules():
    """
    tests a condition where a rule must be deleted
    """
    present_rules = [
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 22),
                ("to_port", 22),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        ),
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 80),
                ("to_port", 80),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        ),
    ]
    desired_rules = [
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 22),
                ("to_port", 22),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        )
    ]
    # can also use: rules_to_delete = [rule for rule in present_rules if rule not in desired_rules]
    rules_to_delete = [
        OrderedDict(
            [
                ("ip_protocol", "tcp"),
                ("from_port", 80),
                ("to_port", 80),
                ("cidr_ip", "0.0.0.0/0"),
            ]
        )
    ]
    assert boto_secgroup._get_rule_changes(desired_rules, present_rules) == (
        rules_to_delete,
        [],
    )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.modjk as modjk


@pytest.fixture
def list_not_str():
    return "workers should be a list not a <class 'str'>"


def test_worker_stopped(list_not_str):
    """
    Test to stop all the workers in the modjk load balancer
    """
    name = "loadbalancer"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    ret.update({"comment": list_not_str})
    assert modjk.worker_stopped(name, "app1") == ret


def test_worker_activated(list_not_str):
    """
    Test to activate all the workers in the modjk load balancer
    """
    name = "loadbalancer"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    ret.update({"comment": list_not_str})
    assert modjk.worker_activated(name, "app1") == ret


def test_worker_disabled(list_not_str):
    """
    Test to disable all the workers in the modjk load balancer
    """
    name = "loadbalancer"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    ret.update({"comment": list_not_str})
    assert modjk.worker_disabled(name, "app1") == ret


def test_worker_recover(list_not_str):
    """
    Test to recover all the workers in the modjk load balancer
    """
    name = "loadbalancer"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    ret.update({"comment": list_not_str})
    assert modjk.worker_recover(name, "app1") == ret



"""
Vultr Cloud Module using python-vultr bindings
==============================================

.. versionadded:: 2016.3.0

The Vultr cloud module is used to control access to the Vultr VPS system.

Use of this module only requires the ``api_key`` parameter.

Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/vultr.conf``:

.. code-block:: yaml

    my-vultr-config:
      # Vultr account api key
      api_key: <supersecretapi_key>
      driver: vultr

Set up the cloud profile at ``/etc/salt/cloud.profiles`` or
``/etc/salt/cloud.profiles.d/vultr.conf``:

.. code-block:: yaml

    nyc-4gb-4cpu-ubuntu-14-04:
      location: 1
      provider: my-vultr-config
      image: 160
      size: 95
      enable_private_network: True

This driver also supports Vultr's `startup script` feature.  You can list startup
scripts in your account with

.. code-block:: bash

    salt-cloud -f list_scripts <name of vultr provider>

That list will include the IDs of the scripts in your account.  Thus, if you
have a script called 'setup-networking' with an ID of 493234 you can specify
that startup script in a profile like so:

.. code-block:: yaml

    nyc-2gb-1cpu-ubuntu-17-04:
      location: 1
      provider: my-vultr-config
      image: 223
      size: 13
      startup_script_id: 493234

Similarly you can also specify a fiewall group ID using the option firewall_group_id. You can list
firewall groups with

.. code-block:: bash

    salt-cloud -f list_firewall_groups <name of vultr provider>

To specify SSH keys to be preinstalled on the server, use the ssh_key_names setting

.. code-block:: yaml

    nyc-2gb-1cpu-ubuntu-17-04:
      location: 1
      provider: my-vultr-config
      image: 223
      size: 13
      ssh_key_names: dev1,dev2,salt-master

You can list SSH keys available on your account using

.. code-block:: bash

    salt-cloud -f list_keypairs <name of vultr provider>

"""

import logging
import pprint
import time
import urllib.parse

import salt.config as config
from salt.exceptions import SaltCloudConfigError, SaltCloudSystemExit

# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "vultr"

DETAILS = {}


def __virtual__():
    """
    Set up the Vultr functions and check for configurations
    """
    if get_configured_provider() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or "vultr", ("api_key",)
    )


def _cache_provider_details(conn=None):
    """
    Provide a place to hang onto results of --list-[locations|sizes|images]
    so we don't have to go out to the API and get them every time.
    """
    DETAILS["avail_locations"] = {}
    DETAILS["avail_sizes"] = {}
    DETAILS["avail_images"] = {}
    locations = avail_locations(conn)
    images = avail_images(conn)
    sizes = avail_sizes(conn)

    for key, location in locations.items():
        DETAILS["avail_locations"][location["name"]] = location
        DETAILS["avail_locations"][key] = location

    for key, image in images.items():
        DETAILS["avail_images"][image["name"]] = image
        DETAILS["avail_images"][key] = image

    for key, vm_size in sizes.items():
        DETAILS["avail_sizes"][vm_size["name"]] = vm_size
        DETAILS["avail_sizes"][key] = vm_size


def avail_locations(conn=None):
    """
    return available datacenter locations
    """
    return _query("regions/list")


def avail_scripts(conn=None):
    """
    return available startup scripts
    """
    return _query("startupscript/list")


def avail_firewall_groups(conn=None):
    """
    return available firewall groups
    """
    return _query("firewall/group_list")


def avail_keys(conn=None):
    """
    return available SSH keys
    """
    return _query("sshkey/list")


def list_scripts(conn=None, call=None):
    """
    return list of Startup Scripts
    """
    return avail_scripts()


def list_firewall_groups(conn=None, call=None):
    """
    return list of firewall groups
    """
    return avail_firewall_groups()


def list_keypairs(conn=None, call=None):
    """
    return list of SSH keys
    """
    return avail_keys()


def show_keypair(kwargs=None, call=None):
    """
    return list of SSH keys
    """
    if not kwargs:
        kwargs = {}

    if "keyname" not in kwargs:
        log.error("A keyname is required.")
        return False

    keys = list_keypairs(call="function")
    keyid = keys[kwargs["keyname"]]["SSHKEYID"]
    log.debug("Key ID is %s", keyid)

    return keys[kwargs["keyname"]]


def avail_sizes(conn=None):
    """
    Return available sizes ("plans" in VultrSpeak)
    """
    return _query("plans/list")


def avail_images(conn=None):
    """
    Return available images
    """
    return _query("os/list")


def list_nodes(**kwargs):
    """
    Return basic data on nodes
    """
    ret = {}

    nodes = list_nodes_full()
    for node in nodes:
        ret[node] = {}
        for prop in "id", "image", "size", "state", "private_ips", "public_ips":
            ret[node][prop] = nodes[node][prop]

    return ret


def list_nodes_full(**kwargs):
    """
    Return all data on nodes
    """
    nodes = _query("server/list")
    ret = {}

    for node in nodes:
        name = nodes[node]["label"]
        ret[name] = nodes[node].copy()
        ret[name]["id"] = node
        ret[name]["image"] = nodes[node]["os"]
        ret[name]["size"] = nodes[node]["VPSPLANID"]
        ret[name]["state"] = nodes[node]["status"]
        ret[name]["private_ips"] = nodes[node]["internal_ip"]
        ret[name]["public_ips"] = nodes[node]["main_ip"]

    return ret


def list_nodes_select(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider, with select fields
    """
    return __utils__["cloud.list_nodes_select"](
        list_nodes_full(),
        __opts__["query.selection"],
        call,
    )


def destroy(name):
    """
    Remove a node from Vultr
    """
    node = show_instance(name, call="action")
    params = {"SUBID": node["SUBID"]}
    result = _query(
        "server/destroy",
        method="POST",
        decode=False,
        data=urllib.parse.urlencode(params),
    )

    # The return of a destroy call is empty in the case of a success.
    # Errors are only indicated via HTTP status code. Status code 200
    # effetively therefore means "success".
    if result.get("body") == "" and result.get("text") == "":
        return True
    return result


def stop(*args, **kwargs):
    """
    Execute a "stop" action on a VM
    """
    return _query("server/halt")


def start(*args, **kwargs):
    """
    Execute a "start" action on a VM
    """
    return _query("server/start")


def show_instance(name, call=None):
    """
    Show the details from the provider concerning an instance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    nodes = list_nodes_full()
    # Find under which cloud service the name is listed, if any
    if name not in nodes:
        return {}
    __utils__["cloud.cache_node"](nodes[name], _get_active_provider_name(), __opts__)
    return nodes[name]


def _lookup_vultrid(which_key, availkey, keyname):
    """
    Helper function to retrieve a Vultr ID
    """
    if DETAILS == {}:
        _cache_provider_details()

    which_key = str(which_key)
    try:
        return DETAILS[availkey][which_key][keyname]
    except KeyError:
        return False


def create(vm_):
    """
    Create a single VM from a data dict
    """
    if "driver" not in vm_:
        vm_["driver"] = vm_["provider"]

    private_networking = config.get_cloud_config_value(
        "enable_private_network",
        vm_,
        __opts__,
        search_global=False,
        default=False,
    )

    ssh_key_ids = config.get_cloud_config_value(
        "ssh_key_names", vm_, __opts__, search_global=False, default=None
    )

    startup_script = config.get_cloud_config_value(
        "startup_script_id",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if startup_script and str(startup_script) not in avail_scripts():
        log.error(
            "Your Vultr account does not have a startup script with ID %s",
            str(startup_script),
        )
        return False

    firewall_group_id = config.get_cloud_config_value(
        "firewall_group_id",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if firewall_group_id and str(firewall_group_id) not in avail_firewall_groups():
        log.error(
            "Your Vultr account does not have a firewall group with ID %s",
            str(firewall_group_id),
        )
        return False
    if ssh_key_ids is not None:
        key_list = ssh_key_ids.split(",")
        available_keys = avail_keys()
        for key in key_list:
            if key and str(key) not in available_keys:
                log.error("Your Vultr account does not have a key with ID %s", str(key))
                return False

    if private_networking is not None:
        if not isinstance(private_networking, bool):
            raise SaltCloudConfigError(
                "'private_networking' should be a boolean value."
            )
    if private_networking is True:
        enable_private_network = "yes"
    else:
        enable_private_network = "no"

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    osid = _lookup_vultrid(vm_["image"], "avail_images", "OSID")
    if not osid:
        log.error("Vultr does not have an image with id or name %s", vm_["image"])
        return False

    vpsplanid = _lookup_vultrid(vm_["size"], "avail_sizes", "VPSPLANID")
    if not vpsplanid:
        log.error("Vultr does not have a size with id or name %s", vm_["size"])
        return False

    dcid = _lookup_vultrid(vm_["location"], "avail_locations", "DCID")
    if not dcid:
        log.error("Vultr does not have a location with id or name %s", vm_["location"])
        return False

    kwargs = {
        "label": vm_["name"],
        "OSID": osid,
        "VPSPLANID": vpsplanid,
        "DCID": dcid,
        "hostname": vm_["name"],
        "enable_private_network": enable_private_network,
    }
    if startup_script:
        kwargs["SCRIPTID"] = startup_script

    if firewall_group_id:
        kwargs["FIREWALLGROUPID"] = firewall_group_id

    if ssh_key_ids:
        kwargs["SSHKEYID"] = ssh_key_ids

    log.info("Creating Cloud VM %s", vm_["name"])

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args={
            "kwargs": __utils__["cloud.filter_event"](
                "requesting", kwargs, list(kwargs)
            ),
        },
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    try:
        data = _query(
            "server/create", method="POST", data=urllib.parse.urlencode(kwargs)
        )
        if int(data.get("status", "200")) >= 300:
            log.error(
                "Error creating %s on Vultr\n\nVultr API returned %s\n",
                vm_["name"],
                data,
            )
            log.error(
                "Status 412 may mean that you are requesting an\n"
                "invalid location, image, or size."
            )

            __utils__["cloud.fire_event"](
                "event",
                "instance request failed",
                "salt/cloud/{}/requesting/failed".format(vm_["name"]),
                args={"kwargs": kwargs},
                sock_dir=__opts__["sock_dir"],
                transport=__opts__["transport"],
            )
            return False
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error creating %s on Vultr\n\n"
            "The following exception was thrown when trying to "
            "run the initial deployment:\n%s",
            vm_["name"],
            exc,
            # Show the traceback if the debug logging level is enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        __utils__["cloud.fire_event"](
            "event",
            "instance request failed",
            "salt/cloud/{}/requesting/failed".format(vm_["name"]),
            args={"kwargs": kwargs},
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )
        return False

    def wait_for_hostname():
        """
        Wait for the IP address to become available
        """
        data = show_instance(vm_["name"], call="action")
        main_ip = str(data.get("main_ip", "0"))
        if main_ip.startswith("0"):
            time.sleep(3)
            return False
        return data["main_ip"]

    def wait_for_default_password():
        """
        Wait for the IP address to become available
        """
        data = show_instance(vm_["name"], call="action")
        # print("Waiting for default password")
        # pprint.pprint(data)
        default_password = str(data.get("default_password", ""))
        if default_password == "" or default_password == "not supported":
            time.sleep(1)
            return False
        return data["default_password"]

    def wait_for_status():
        """
        Wait for the IP address to become available
        """
        data = show_instance(vm_["name"], call="action")
        # print("Waiting for status normal")
        # pprint.pprint(data)
        if str(data.get("status", "")) != "active":
            time.sleep(1)
            return False
        return data["default_password"]

    def wait_for_server_state():
        """
        Wait for the IP address to become available
        """
        data = show_instance(vm_["name"], call="action")
        # print("Waiting for server state ok")
        # pprint.pprint(data)
        if str(data.get("server_state", "")) != "ok":
            time.sleep(1)
            return False
        return data["default_password"]

    vm_["ssh_host"] = __utils__["cloud.wait_for_fun"](
        wait_for_hostname,
        timeout=config.get_cloud_config_value(
            "wait_for_fun_timeout", vm_, __opts__, default=15 * 60
        ),
    )
    vm_["password"] = __utils__["cloud.wait_for_fun"](
        wait_for_default_password,
        timeout=config.get_cloud_config_value(
            "wait_for_fun_timeout", vm_, __opts__, default=15 * 60
        ),
    )
    __utils__["cloud.wait_for_fun"](
        wait_for_status,
        timeout=config.get_cloud_config_value(
            "wait_for_fun_timeout", vm_, __opts__, default=15 * 60
        ),
    )
    __utils__["cloud.wait_for_fun"](
        wait_for_server_state,
        timeout=config.get_cloud_config_value(
            "wait_for_fun_timeout", vm_, __opts__, default=15 * 60
        ),
    )

    __opts__["hard_timeout"] = config.get_cloud_config_value(
        "hard_timeout",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default=None,
    )

    # Bootstrap
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)

    ret.update(show_instance(vm_["name"], call="action"))

    log.info("Created Cloud VM '%s'", vm_["name"])
    log.debug("'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def _query(path, method="GET", data=None, params=None, header_dict=None, decode=True):
    """
    Perform a query directly against the Vultr REST API
    """
    api_key = config.get_cloud_config_value(
        "api_key",
        get_configured_provider(),
        __opts__,
        search_global=False,
    )
    management_host = config.get_cloud_config_value(
        "management_host",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default="api.vultr.com",
    )
    url = "https://{management_host}/v1/{path}?api_key={api_key}".format(
        management_host=management_host,
        path=path,
        api_key=api_key,
    )

    if header_dict is None:
        header_dict = {}

    result = __utils__["http.query"](
        url,
        method=method,
        params=params,
        data=data,
        header_dict=header_dict,
        port=443,
        text=True,
        decode=decode,
        decode_type="json",
        hide_fields=["api_key"],
        opts=__opts__,
    )
    if "dict" in result:
        return result["dict"]

    return result



"""Bridges between the `asyncio` module and Tornado IOLoop.

.. versionadded:: 3.2

This module integrates Tornado with the ``asyncio`` module introduced
in Python 3.4 (and available `as a separate download
<https://pypi.python.org/pypi/asyncio>`_ for Python 3.3).  This makes
it possible to combine the two libraries on the same event loop.

Most applications should use `AsyncIOMainLoop` to run Tornado on the
default ``asyncio`` event loop.  Applications that need to run event
loops on multiple threads may use `AsyncIOLoop` to create multiple
loops.

.. note::

   Tornado requires the `~asyncio.AbstractEventLoop.add_reader` family of
   methods, so it is not compatible with the `~asyncio.ProactorEventLoop` on
   Windows. Use the `~asyncio.SelectorEventLoop` instead.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function
import functools

import salt.ext.tornado.concurrent
from salt.ext.tornado.gen import convert_yielded
from salt.ext.tornado.ioloop import IOLoop
from salt.ext.tornado import stack_context

try:
    # Import the real asyncio module for py33+ first.  Older versions of the
    # trollius backport also use this name.
    import asyncio  # type: ignore
except ImportError as e:
    # Asyncio itself isn't available; see if trollius is (backport to py26+).
    try:
        import trollius as asyncio  # type: ignore
    except ImportError:
        # Re-raise the original asyncio error, not the trollius one.
        raise e


class BaseAsyncIOLoop(IOLoop):
    def initialize(self, asyncio_loop, close_loop=False, **kwargs):
        super(BaseAsyncIOLoop, self).initialize(**kwargs)
        self.asyncio_loop = asyncio_loop
        self.close_loop = close_loop
        # Maps fd to (fileobj, handler function) pair (as in IOLoop.add_handler)
        self.handlers = {}
        # Set of fds listening for reads/writes
        self.readers = set()
        self.writers = set()
        self.closing = False

    def close(self, all_fds=False):
        self.closing = True
        for fd in list(self.handlers):
            fileobj, handler_func = self.handlers[fd]
            self.remove_handler(fd)
            if all_fds:
                self.close_fd(fileobj)
        if self.close_loop:
            self.asyncio_loop.close()

    def add_handler(self, fd, handler, events):
        fd, fileobj = self.split_fd(fd)
        if fd in self.handlers:
            raise ValueError("fd %s added twice" % fd)
        self.handlers[fd] = (fileobj, stack_context.wrap(handler))
        if events & IOLoop.READ:
            self.asyncio_loop.add_reader(
                fd, self._handle_events, fd, IOLoop.READ)
            self.readers.add(fd)
        if events & IOLoop.WRITE:
            self.asyncio_loop.add_writer(
                fd, self._handle_events, fd, IOLoop.WRITE)
            self.writers.add(fd)

    def update_handler(self, fd, events):
        fd, fileobj = self.split_fd(fd)
        if events & IOLoop.READ:
            if fd not in self.readers:
                self.asyncio_loop.add_reader(
                    fd, self._handle_events, fd, IOLoop.READ)
                self.readers.add(fd)
        else:
            if fd in self.readers:
                self.asyncio_loop.remove_reader(fd)
                self.readers.remove(fd)
        if events & IOLoop.WRITE:
            if fd not in self.writers:
                self.asyncio_loop.add_writer(
                    fd, self._handle_events, fd, IOLoop.WRITE)
                self.writers.add(fd)
        else:
            if fd in self.writers:
                self.asyncio_loop.remove_writer(fd)
                self.writers.remove(fd)

    def remove_handler(self, fd):
        fd, fileobj = self.split_fd(fd)
        if fd not in self.handlers:
            return
        if fd in self.readers:
            self.asyncio_loop.remove_reader(fd)
            self.readers.remove(fd)
        if fd in self.writers:
            self.asyncio_loop.remove_writer(fd)
            self.writers.remove(fd)
        del self.handlers[fd]

    def _handle_events(self, fd, events):
        fileobj, handler_func = self.handlers[fd]
        handler_func(fileobj, events)

    def start(self):
        old_current = IOLoop.current(instance=False)
        try:
            self._setup_logging()
            self.make_current()
            self.asyncio_loop.run_forever()
        finally:
            if old_current is None:
                IOLoop.clear_current()
            else:
                old_current.make_current()

    def stop(self):
        self.asyncio_loop.stop()

    def call_at(self, when, callback, *args, **kwargs):
        # asyncio.call_at supports *args but not **kwargs, so bind them here.
        # We do not synchronize self.time and asyncio_loop.time, so
        # convert from absolute to relative.
        return self.asyncio_loop.call_later(
            max(0, when - self.time()), self._run_callback,
            functools.partial(stack_context.wrap(callback), *args, **kwargs))

    def remove_timeout(self, timeout):
        timeout.cancel()

    def add_callback(self, callback, *args, **kwargs):
        if self.closing:
            # TODO: this is racy; we need a lock to ensure that the
            # loop isn't closed during call_soon_threadsafe.
            raise RuntimeError("IOLoop is closing")
        self.asyncio_loop.call_soon_threadsafe(
            self._run_callback,
            functools.partial(stack_context.wrap(callback), *args, **kwargs))

    add_callback_from_signal = add_callback


class AsyncIOMainLoop(BaseAsyncIOLoop):
    """``AsyncIOMainLoop`` creates an `.IOLoop` that corresponds to the
    current ``asyncio`` event loop (i.e. the one returned by
    ``asyncio.get_event_loop()``).  Recommended usage::

        from salt.ext.tornado.platform.asyncio import AsyncIOMainLoop
        import asyncio
        AsyncIOMainLoop().install()
        asyncio.get_event_loop().run_forever()

    See also :meth:`tornado.ioloop.IOLoop.install` for general notes on
    installing alternative IOLoops.
    """
    def initialize(self, **kwargs):
        super(AsyncIOMainLoop, self).initialize(asyncio.get_event_loop(),
                                                close_loop=False, **kwargs)


class AsyncIOLoop(BaseAsyncIOLoop):
    """``AsyncIOLoop`` is an `.IOLoop` that runs on an ``asyncio`` event loop.
    This class follows the usual Tornado semantics for creating new
    ``IOLoops``; these loops are not necessarily related to the
    ``asyncio`` default event loop.  Recommended usage::

        from salt.ext.tornado.ioloop import IOLoop
        IOLoop.configure('tornado.platform.asyncio.AsyncIOLoop')
        IOLoop.current().start()

    Each ``AsyncIOLoop`` creates a new ``asyncio.EventLoop``; this object
    can be accessed with the ``asyncio_loop`` attribute.
    """
    def initialize(self, **kwargs):
        loop = asyncio.new_event_loop()
        try:
            super(AsyncIOLoop, self).initialize(loop, close_loop=True, **kwargs)
        except Exception:
            # If initialize() does not succeed (taking ownership of the loop),
            # we have to close it.
            loop.close()
            raise


def to_tornado_future(asyncio_future):
    """Convert an `asyncio.Future` to a `tornado.concurrent.Future`.

    .. versionadded:: 4.1
    """
    tf = salt.ext.tornado.concurrent.Future()
    salt.ext.tornado.concurrent.chain_future(asyncio_future, tf)
    return tf


def to_asyncio_future(tornado_future):
    """Convert a Tornado yieldable object to an `asyncio.Future`.

    .. versionadded:: 4.1

    .. versionchanged:: 4.3
       Now accepts any yieldable object, not just
       `tornado.concurrent.Future`.
    """
    tornado_future = convert_yielded(tornado_future)
    af = asyncio.Future()
    salt.ext.tornado.concurrent.chain_future(tornado_future, af)
    return af


if hasattr(convert_yielded, 'register'):
    convert_yielded.register(asyncio.Future, to_tornado_future)  # type: ignore



import pytest

from salt.utils.dictdiffer import RecursiveDictDiffer
from salt.utils.listdiffer import list_diff


@pytest.fixture
def get_old_list():
    return [
        {"key": 1, "value": "foo1", "int_value": 101},
        {"key": 2, "value": "foo2", "int_value": 102},
        {"key": 3, "value": "foo3", "int_value": 103},
    ]


@pytest.fixture
def get_new_list():
    return [
        {"key": 1, "value": "foo1", "int_value": 101},
        {"key": 2, "value": "foo2", "int_value": 112},
        {"key": 5, "value": "foo5", "int_value": 105},
    ]


@pytest.fixture
def get_list_diff(get_old_list, get_new_list):
    return list_diff(get_old_list, get_new_list, key="key")


def test_added(get_list_diff):
    assert len(get_list_diff.added) == 1
    assert get_list_diff.added[0] == {"key": 5, "value": "foo5", "int_value": 105}


def test_removed(get_list_diff):
    assert len(get_list_diff.removed) == 1
    assert get_list_diff.removed[0] == {"key": 3, "value": "foo3", "int_value": 103}


def test_diffs(get_list_diff):
    assert len(get_list_diff.diffs) == 3
    assert get_list_diff.diffs[0] == {2: {"int_value": {"new": 112, "old": 102}}}

    # Added items
    assert get_list_diff.diffs[1] == {
        5: {
            "int_value": {"new": 105, "old": RecursiveDictDiffer.NONE_VALUE},
            "key": {"new": 5, "old": RecursiveDictDiffer.NONE_VALUE},
            "value": {"new": "foo5", "old": RecursiveDictDiffer.NONE_VALUE},
        }
    }

    # Removed items
    assert get_list_diff.diffs[2] == {
        3: {
            "int_value": {"new": RecursiveDictDiffer.NONE_VALUE, "old": 103},
            "key": {"new": RecursiveDictDiffer.NONE_VALUE, "old": 3},
            "value": {"new": RecursiveDictDiffer.NONE_VALUE, "old": "foo3"},
        }
    }


def test_new_values(get_list_diff):
    assert len(get_list_diff.new_values) == 2
    assert get_list_diff.new_values[0] == {"key": 2, "int_value": 112}
    assert get_list_diff.new_values[1] == {"key": 5, "value": "foo5", "int_value": 105}


def test_old_values(get_list_diff):
    assert len(get_list_diff.old_values) == 2
    assert get_list_diff.old_values[0] == {"key": 2, "int_value": 102}
    assert get_list_diff.old_values[1] == {"key": 3, "value": "foo3", "int_value": 103}


def test_changed_all(get_list_diff):
    assert get_list_diff.changed(selection="all") == [
        "key.2.int_value",
        "key.5.int_value",
        "key.5.value",
        "key.3.int_value",
        "key.3.value",
    ]


def test_changed_intersect(get_list_diff):
    assert get_list_diff.changed(selection="intersect") == ["key.2.int_value"]


def test_changes_str(get_list_diff):
    expected = """\tidentified by key 2:
\tint_value from 102 to 112
\tidentified by key 3:
\twill be removed
\tidentified by key 5:
\twill be added
"""
    assert get_list_diff.changes_str == expected


def test_intersect(get_list_diff):
    expected = [
        {
            "key": 1,
            "old": {"key": 1, "value": "foo1", "int_value": 101},
            "new": {"key": 1, "value": "foo1", "int_value": 101},
        },
        {
            "key": 2,
            "old": {"key": 2, "value": "foo2", "int_value": 102},
            "new": {"key": 2, "value": "foo2", "int_value": 112},
        },
    ]
    test_isect = get_list_diff.intersect
    assert test_isect == expected


def test_remove_diff_intersect(get_list_diff):
    expected = [
        {
            "key": 1,
            "old": {"key": 1, "int_value": 101},
            "new": {"key": 1, "int_value": 101},
        },
        {
            "key": 2,
            "old": {"key": 2, "int_value": 102},
            "new": {"key": 2, "int_value": 112},
        },
    ]

    get_list_diff.remove_diff(diff_key="value")
    test_isect = get_list_diff.intersect
    assert test_isect == expected


def test_remove_diff_removed(get_list_diff):
    expected = [
        {
            "key": 1,
            "old": {"key": 1, "value": "foo1", "int_value": 101},
            "new": {"key": 1, "value": "foo1", "int_value": 101},
        },
        {
            "key": 2,
            "old": {"key": 2, "value": "foo2", "int_value": 102},
            "new": {"key": 2, "value": "foo2", "int_value": 112},
        },
    ]
    get_list_diff.remove_diff(diff_key="value", diff_list="removed")
    test_isect = get_list_diff.intersect
    assert test_isect == expected


def test_changes_str2(get_list_diff):
    expected = """  key=2 (updated):
    int_value from 102 to 112
  key=3 (removed)
  key=5 (added): {'key': 5, 'value': 'foo5', 'int_value': 105}"""
    test_changes = get_list_diff.changes_str2
    assert test_changes == expected


def test_current_list(get_list_diff):
    expected = [
        {"key": 1, "value": "foo1", "int_value": 101},
        {"key": 2, "value": "foo2", "int_value": 102},
        {"key": 3, "value": "foo3", "int_value": 103},
    ]
    test_curr_list = get_list_diff.current_list
    assert test_curr_list == expected


def test_new_list(get_list_diff):
    expected = [
        {"key": 1, "value": "foo1", "int_value": 101},
        {"key": 2, "value": "foo2", "int_value": 112},
        {"key": 5, "value": "foo5", "int_value": 105},
    ]
    test_new_list = get_list_diff.new_list
    assert test_new_list == expected



import logging
import os
import shutil
import subprocess
import time

import pytest
from pytestshellutils.exceptions import FactoryTimeout

import salt.utils.platform
from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def _salt_mm_master_1(request, salt_factories):
    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "127.0.0.1",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
        "log_granular_levels": {
            "salt": "info",
            "salt.transport": "debug",
            "salt.channel": "debug",
            "salt.utils.event": "debug",
        },
    }
    factory = salt_factories.salt_master_daemon(
        "mm-master-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    # Start the factory so the key files will be generated. After, we'll yeild
    # the factory and the deamon will not be running.
    with factory.started(start_timeout=120):
        pass
    yield factory


@pytest.fixture
def salt_mm_master_1(_salt_mm_master_1):
    with _salt_mm_master_1.started(start_timeout=120):
        yield _salt_mm_master_1


@pytest.fixture
def mm_master_1_salt_cli(salt_mm_master_1):
    return salt_mm_master_1.salt_cli(timeout=120)


@pytest.fixture(scope="package")
def _salt_mm_master_2(salt_factories, _salt_mm_master_1):
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
        subprocess.check_output(["ifconfig", "lo0", "alias", "127.0.0.2", "up"])

    config_defaults = {
        "open_mode": True,
        "transport": _salt_mm_master_1.config["transport"],
    }
    config_overrides = {
        "interface": "127.0.0.2",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
        "log_granular_levels": {
            "salt": "info",
            "salt.transport": "debug",
            "salt.channel": "debug",
            "salt.utils.event": "debug",
        },
    }

    # Use the same ports for both masters, they are binding to different interfaces
    for key in (
        "ret_port",
        "publish_port",
    ):
        config_overrides[key] = _salt_mm_master_1.config[key]
    factory = salt_factories.salt_master_daemon(
        "mm-master-2",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )

    # The secondary salt master depends on the primarily salt master fixture
    # because we need to clone the keys
    for keyfile in ("master.pem", "master.pub"):
        shutil.copyfile(
            os.path.join(_salt_mm_master_1.config["pki_dir"], keyfile),
            os.path.join(factory.config["pki_dir"], keyfile),
        )
    yield factory


@pytest.fixture
def salt_mm_master_2(_salt_mm_master_2):
    with _salt_mm_master_2.started(start_timeout=120):
        yield _salt_mm_master_2


@pytest.fixture
def mm_master_2_salt_cli(salt_mm_master_2):
    return salt_mm_master_2.salt_cli(timeout=120)


@pytest.fixture(scope="package")
def _salt_mm_minion_1(_salt_mm_master_1, _salt_mm_master_2):
    config_defaults = {
        "transport": _salt_mm_master_1.config["transport"],
    }

    mm_master_1_port = _salt_mm_master_1.config["ret_port"]
    mm_master_1_addr = _salt_mm_master_1.config["interface"]
    mm_master_2_port = _salt_mm_master_2.config["ret_port"]
    mm_master_2_addr = _salt_mm_master_2.config["interface"]
    config_overrides = {
        "master": [
            f"{mm_master_1_addr}:{mm_master_1_port}",
            f"{mm_master_2_addr}:{mm_master_2_port}",
        ],
        "test.foo": "baz",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        "log_granular_levels": {
            "salt": "info",
            "salt.minion": "debug",
            "salt.transport": "debug",
            "salt.channel": "debug",
            "salt.utils.event": "debug",
        },
    }
    yield _salt_mm_master_1.salt_minion_daemon(
        "mm-minion-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )


@pytest.fixture
def salt_mm_minion_1(_salt_mm_minion_1, salt_mm_master_1, salt_mm_master_2):
    with _salt_mm_minion_1.started(start_timeout=120):
        yield _salt_mm_minion_1


@pytest.fixture(scope="package")
def _salt_mm_minion_2(_salt_mm_master_1, _salt_mm_master_2):
    config_defaults = {
        "transport": _salt_mm_master_1.config["transport"],
    }

    mm_master_1_port = _salt_mm_master_1.config["ret_port"]
    mm_master_1_addr = _salt_mm_master_1.config["interface"]
    mm_master_2_port = _salt_mm_master_2.config["ret_port"]
    mm_master_2_addr = _salt_mm_master_2.config["interface"]
    config_overrides = {
        "master": [
            f"{mm_master_1_addr}:{mm_master_1_port}",
            f"{mm_master_2_addr}:{mm_master_2_port}",
        ],
        "test.foo": "baz",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        "log_granular_levels": {
            "salt": "info",
            "salt.minion": "debug",
            "salt.transport": "debug",
            "salt.channel": "debug",
            "salt.utils.event": "debug",
        },
    }
    yield _salt_mm_master_2.salt_minion_daemon(
        "mm-minion-2",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )


@pytest.fixture
def salt_mm_minion_2(_salt_mm_minion_2, salt_mm_master_1, salt_mm_master_2):
    with _salt_mm_minion_2.started(start_timeout=120):
        yield _salt_mm_minion_2


@pytest.fixture
def run_salt_cmds():
    def _run_salt_cmds_fn(clis, minions):
        """
        Run test.ping from all clis to all minions
        """
        returned_minions = []
        minion_instances = {minion.id: minion for minion in minions}
        clis_to_check = {minion.id: list(clis) for minion in minions}

        attempts = 6
        timeout = 5
        if salt.utils.platform.spawning_platform():
            timeout *= 2
        while attempts:
            if not clis_to_check:
                break
            for minion in list(clis_to_check):
                if not clis_to_check[minion]:
                    clis_to_check.pop(minion)
                    continue
                for cli in list(clis_to_check[minion]):
                    try:
                        ret = cli.run(
                            f"--timeout={timeout}",
                            "test.ping",
                            minion_tgt=minion,
                            _timeout=2 * timeout,
                        )
                        if ret.returncode == 0 and ret.data is True:
                            returned_minions.append((cli, minion_instances[minion]))
                            clis_to_check[minion].remove(cli)
                    except FactoryTimeout:
                        log.debug(
                            "Failed to execute test.ping from %s to %s.",
                            cli.get_display_name(),
                            minion,
                        )
            time.sleep(1)
            attempts -= 1

        return returned_minions

    return _run_salt_cmds_fn


@pytest.fixture
def ensure_connections(
    salt_mm_minion_1,
    salt_mm_minion_2,
    mm_master_1_salt_cli,
    mm_master_2_salt_cli,
    run_salt_cmds,
):
    # define the function
    def _ensure_connections_fn(clis, minions, *, timeout=120, interval=5):
        """
        Verify every CLI/minion pair can talk to each other.

        On slower transports (for example TCP on ARM builders) the minions can
        take longer than the original 30 seconds to re-auth.  Instead of a
        fixed number of retries, spin until ``timeout`` elapses so the caller
        only fails when we truly give up.
        """

        deadline = time.monotonic() + timeout
        # Track the expected pairings as a set of human-readable identifiers
        expected = {
            (cli.get_display_name(), minion.id) for cli in clis for minion in minions
        }
        last_returned = []

        log.debug("ensure_connections start: expected=%s", expected)
        while time.monotonic() < deadline:
            returned = run_salt_cmds(clis, minions)
            returned_pairs = {
                (cli.get_display_name(), minion.id) for cli, minion in returned
            }
            if returned_pairs == expected:
                log.debug("ensure_connections successful: returned=%s", returned_pairs)
                return

            # Remember the last good attempt for debugging
            last_returned = list(returned_pairs)
            missing = sorted(expected - returned_pairs)
            log.debug(
                "ensure_connections retrying; missing=%s interval=%ss",
                missing,
                interval,
            )
            time.sleep(interval)

        missing = sorted(expected - set(last_returned))
        pytest.fail(
            "Could not ensure the connections were okay. "
            f"Missing responses for {missing} after {timeout}s."
        )

    # run the function to ensure initial connections
    _ensure_connections_fn(
        [mm_master_1_salt_cli, mm_master_2_salt_cli],
        [salt_mm_minion_1, salt_mm_minion_2],
    )

    # Give this function back for further use in test fn bodies
    return _ensure_connections_fn
